from dataclasses import is_dataclass
from enum import Enum
from typing import (
    Any,
    Callable,
    Collection,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
)

from apischema import settings
from apischema.aliases import Aliaser
from apischema.cache import cache
from apischema.conversions.conversions import (
    Conversions,
    ResolvedConversion,
    handle_container_conversions,
    resolve_conversions,
)
from apischema.conversions.dataclass_models import DataclassModel
from apischema.conversions.visitor import SerializationVisitor
from apischema.dataclass_utils import (
    dataclass_types_and_fields,
    get_alias,
    get_field_conversions,
)
from apischema.fields import FIELDS_SET_ATTR, fields_set
from apischema.metadata.keys import SKIP_METADATA, check_metadata, is_aggregate_field
from apischema.serialization.serialized_methods import get_serialized_methods
from apischema.types import COLLECTION_TYPES, MAPPING_TYPES, PRIMITIVE_TYPES
from apischema.utils import OperationKind, Undefined
from apischema.visitor import Unsupported

PRIMITIVE_TYPES_SET = set(PRIMITIVE_TYPES)
COLLECTION_TYPE_SET = set(COLLECTION_TYPES)
MAPPING_TYPE_SET = set(MAPPING_TYPES)


@cache
def serialization_fields(
    cls: Type, aliaser: Aliaser
) -> Tuple[
    Sequence[Tuple[str, str, Optional[Conversions]]],
    Sequence[Tuple[str, Optional[Conversions]]],
    Sequence[Tuple[Callable[[Any], Any], str, Optional[Conversions]]],
]:
    types, fields, _ = dataclass_types_and_fields(cls)  # type: ignore
    normal_fields, aggregate_fields, serialized_fields = [], [], []
    for field in fields:
        if SKIP_METADATA in field.metadata:
            continue
        check_metadata(field)
        conversions = get_field_conversions(field, OperationKind.SERIALIZATION)
        if is_aggregate_field(field):
            aggregate_fields.append((field.name, conversions))
        else:
            normal_fields.append((field.name, aliaser(get_alias(field)), conversions))
    for name, (serialized, _) in get_serialized_methods(cls).items():
        serialized_fields.append(
            (serialized.func, aliaser(name), serialized.conversions)
        )

    return normal_fields, aggregate_fields, serialized_fields


@cache
def get_conversions(
    tp: Type, conversions: Optional[Conversions]
) -> Tuple[Optional[ResolvedConversion], bool]:
    return SerializationVisitor.get_conversions(tp, resolve_conversions(conversions))


def serialize(
    obj: Any,
    *,
    conversions: Conversions = None,
    aliaser: Aliaser = None,
    exclude_unset: bool = None,
) -> Any:
    if aliaser is None:
        aliaser = settings.aliaser()
    if exclude_unset is None:
        exclude_unset = settings.exclude_unset

    def _serialize(
        obj: Any,
        exc_unset: bool,
        conversions: Conversions = None,
    ) -> Any:
        assert aliaser is not None
        cls = obj.__class__
        conversion, dynamic = get_conversions(cls, conversions)
        if conversion is not None:
            if conversion.exclude_unset is not None:
                exc_unset = conversion.exclude_unset
            if isinstance(conversion.target, DataclassModel):
                cls = conversion.target.dataclass
            else:
                return _serialize(
                    conversion.converter(obj),  # type: ignore
                    exc_unset,
                    handle_container_conversions(
                        conversion.target,
                        conversion.sub_conversions,
                        conversions,
                        dynamic,
                    ),
                )
        if cls in PRIMITIVE_TYPES_SET:
            return obj
        if cls in COLLECTION_TYPE_SET:
            return [_serialize(elt, exc_unset, conversions) for elt in obj]
        if cls in MAPPING_TYPE_SET:
            return {
                _serialize(key, exc_unset, conversions): _serialize(
                    value, exc_unset, conversions
                )
                for key, value in obj.items()
            }
        if is_dataclass(cls):
            fields, aggregate_fields, serialized_fields = serialization_fields(
                cls, aliaser
            )
            if exclude_unset and hasattr(obj, FIELDS_SET_ATTR):
                fields_set_ = fields_set(obj)
                fields = [
                    (name, alias, method)
                    for (name, alias, method) in fields
                    if name in fields_set_
                ]
                aggregate_fields = [
                    (name, method)
                    for (name, method) in aggregate_fields
                    if name in fields_set_
                ]
            result = {}
            # properties before normal fields to avoid overloading a field with property
            for name, conv in aggregate_fields:
                result.update(_serialize(getattr(obj, name), exc_unset, conv))
            for name, alias, conv in fields:
                attr = getattr(obj, name)
                if attr is not Undefined:
                    result[alias] = _serialize(attr, exc_unset, conv)
            for func, alias, conv in serialized_fields:
                res = func(obj)
                if res is not Undefined:
                    result[alias] = _serialize(res, exc_unset, conv)
            return result
        if obj is Undefined:
            raise Unsupported(cls)
        if issubclass(cls, Enum):
            return _serialize(obj.value, exc_unset)
        if isinstance(obj, PRIMITIVE_TYPES):
            return obj
        if isinstance(obj, Mapping):
            return {
                _serialize(key, exc_unset, conversions): _serialize(
                    value, exc_unset, conversions
                )
                for key, value in obj.items()
            }
        if issubclass(cls, tuple) and hasattr(cls, "_fields"):
            result = {}
            for field_name in obj._fields:
                attr = getattr(obj, field_name)
                if attr is not Undefined:
                    result[aliaser(field_name)] = attr
            for name, (serialized, _) in get_serialized_methods(cls).items():
                res = serialized.func(obj)
                if res is not Undefined:
                    result[aliaser(name)] = _serialize(
                        res, exc_unset, serialized.conversions
                    )
            return result
        if isinstance(obj, Collection):
            return [_serialize(elt, exc_unset, conversions) for elt in obj]
        raise Unsupported(cls)

    return _serialize(obj, exclude_unset, conversions)



from contextlib import suppress
from dataclasses import Field
from enum import Enum
from typing import Any, Dict, Iterable, Mapping, Optional, Sequence, Tuple, Type

from apischema.conversions.visitor import ConversionsVisitor
from apischema.dataclass_utils import get_field_conversions, get_fields
from apischema.json_schema.refs import get_ref, schema_ref
from apischema.skip import filter_skipped
from apischema.types import AnyType
from apischema.utils import contains

try:
    from apischema.typing import Annotated
except ImportError:
    Annotated = ...  # type: ignore

Refs = Dict[str, Tuple[AnyType, int]]


class Recursive(Exception):
    pass


class RefsExtractor(ConversionsVisitor):
    def __init__(self, refs: Refs):
        super().__init__()
        self.refs = refs
        self._rec_guard: Dict[AnyType, bool] = {}

    def _incr_ref(self, ref: Optional[str], tp: AnyType) -> bool:
        if ref is None:
            return False
        else:
            ref_cls, count = self.refs.get(ref, (tp, 0))
            if ref_cls != tp:
                raise ValueError(
                    f"Types {tp} and {self.refs[ref][0]} share same reference '{ref}'"
                )
            self.refs[ref] = (ref_cls, count + 1)
            return count > 0

    def annotated(self, tp: AnyType, annotations: Sequence[Any]):
        for annotation in reversed(annotations):
            if isinstance(annotation, schema_ref):
                annotation.check_type(tp)
                ref = annotation.ref
                if not isinstance(ref, str):
                    raise ValueError("Annotated schema_ref can only be str")
                annotated = Annotated[(tp, *annotations)]  # type: ignore
                if self._incr_ref(ref, annotated):
                    return
        return self.visit(tp)

    def any(self):
        pass

    def collection(self, cls: Type[Iterable], value_type: AnyType):
        return self.visit(value_type)

    def dataclass(
        self,
        cls: Type,
        types: Mapping[str, AnyType],
        fields: Sequence[Field],
        init_vars: Sequence[Field],
    ):
        for field in get_fields(fields, init_vars, self.operation):
            self.visit_with_conversions(
                types[field.name], get_field_conversions(field, self.operation)
            )

    def enum(self, cls: Type[Enum]):
        pass

    def literal(self, values: Sequence[Any]):
        pass

    def mapping(self, cls: Type[Mapping], key_type: AnyType, value_type: AnyType):
        self.visit(key_type)
        self.visit(value_type)

    def named_tuple(
        self,
        cls: Type[Tuple],
        types: Mapping[str, AnyType],
        defaults: Mapping[str, Any],
    ):
        for cls in types.values():
            self.visit(cls)

    def new_type(self, tp: AnyType, super_type: AnyType):
        self.visit(super_type)

    def primitive(self, cls: Type):
        pass

    def subprimitive(self, cls: Type, superclass: Type):
        self.visit(superclass)

    def tuple(self, types: Sequence[AnyType]):
        for cls in types:
            self.visit(cls)

    def typed_dict(self, cls: Type, keys: Mapping[str, AnyType], total: bool):
        for cls in keys.values():
            self.visit(cls)

    def _union_result(self, results: Iterable):
        for _ in results:
            pass

    def union(self, alternatives: Sequence[AnyType]):
        for tp in filter_skipped(alternatives, schema_only=True):
            self.visit(tp)

    def visit(self, tp: AnyType):
        dynamic = self._apply_dynamic_conversions(tp)
        ref_tp = dynamic if dynamic is not None else tp
        if not self._incr_ref(get_ref(ref_tp), ref_tp):
            if contains(self._rec_guard, ref_tp) and self._rec_guard[ref_tp]:
                raise TypeError(f"Recursive type {tp} need a ref")
            with suppress(TypeError):
                self._rec_guard[ref_tp] = ref_tp in self._rec_guard
            try:
                super().visit(ref_tp)
            finally:
                with suppress(TypeError):
                    if self._rec_guard[ref_tp]:
                        self._rec_guard[ref_tp] = False
                    else:
                        del self._rec_guard[ref_tp]



from dataclasses import dataclass
from typing import Generic, Optional, Sequence, TypeVar

Cursor_ = TypeVar("Cursor_")
Node_ = TypeVar("Node_")


@dataclass
class Edge(Generic[Node_, Cursor_]):
    node: Node_
    cursor: Cursor_


@dataclass
class PageInfo(Generic[Cursor_]):
    has_previous_page: bool = False
    has_next_page: bool = False
    start_cursor: Optional[Cursor_] = None
    end_cursor: Optional[Cursor_] = None

    @staticmethod
    def from_edges(
        edges: Sequence[Optional[Edge[Node_, Cursor_]]],
        has_previous_page: bool = False,
        has_next_page: bool = False,
    ) -> "PageInfo":
        start_cursor, end_cursor = None, None
        if edges is not None:
            if edges[0] is not None:
                start_cursor = edges[0].cursor
            if edges[-1] is not None:
                end_cursor = edges[-1].cursor
        return PageInfo(has_previous_page, has_next_page, start_cursor, end_cursor)


Edge_ = TypeVar("Edge_", bound=Edge)


@dataclass
class Connection(Generic[Node_, Cursor_, Edge_]):
    edges: Optional[Sequence[Optional[Edge_]]]
    page_info: PageInfo[Cursor_]



__all__ = [
    "Undefined",
    "UndefinedType",
    "Unsupported",
    "ValidationError",
    "alias",
    "deserialization",
    "deserialize",
    "deserializer",
    "properties",
    "reset_cache",
    "schema",
    "schema_ref",
    "serialize",
    "serialized",
    "serializer",
    "validator",
]


from . import (  # noqa: F401
    conversions,
    dataclasses,
    fields,
    json_schema,
    metadata,
    settings,
    skip,
    tagged_unions,
    validation,
)
from .aliases import alias
from .cache import reset_cache
from .conversions import deserializer, serializer
from .deserialization import deserialize
from .json_schema.refs import schema_ref
from .json_schema.schema import schema
from .metadata import properties
from .serialization import serialize
from .serialization.serialized_methods import serialized
from .utils import Undefined, UndefinedType
from .validation import ValidationError, validator
from .visitor import Unsupported

try:
    from . import graphql  # noqa: F401

    __all__.append("graphql")
except ImportError:
    pass


def __getattr__(name):
    if name == "graphql":
        raise AttributeError(
            "GraphQL feature requires graphql-core library\n"
            "Run `pip install apischema[graphql]` to install it"
        )
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


def register_default_conversions():
    """Handle standard library + internal types"""
    from . import std_types  # noqa: F401

    deserializer(ValidationError.deserialize)
    serializer(ValidationError.serialize)


register_default_conversions()



from contextlib import suppress
from dataclasses import dataclass, fields
from typing import (
    Any,
    Dict,
    Mapping,
    Optional,
    Pattern,
    Sequence,
    TypeVar,
    Union,
    overload,
)

from apischema.types import AnyType, MetadataMixin, Number
from apischema.utils import Undefined, contains, merge_opts, replace_builtins
from .annotations import Annotations, Deprecated, merge_annotations
from .constraints import (
    ArrayConstraints,
    Constraints,
    NumberConstraints,
    ObjectConstraints,
    StringConstraints,
    merge_constraints,
)
from apischema.metadata.keys import SCHEMA_METADATA
from apischema.typing import get_origin

try:
    from apischema.typing import Annotated
except ImportError:
    Annotated = ...  # type: ignore

T = TypeVar("T")


@dataclass(frozen=True)
class Schema(MetadataMixin):
    key = SCHEMA_METADATA
    annotations: Optional[Annotations] = None
    constraints: Optional[Constraints] = None
    override: bool = False

    def __call__(self, tp: T) -> T:
        if get_origin(tp) is Annotated:
            raise TypeError("Cannot register schema on Annotated type")
        _schemas[replace_builtins(tp)] = self
        return tp

    def __set_name__(self, owner, name):
        self.__call__(owner)

    def as_dict(self) -> Mapping[str, Any]:
        result: Dict[str, Any] = {}
        if self.constraints is not None:
            result.update(self.constraints.as_dict())
        if self.annotations is not None:
            result.update(self.annotations.as_dict())
        return result


_annotations_fields = {f.name for f in fields(Annotations)}


@overload
def schema(
    *,
    title: Optional[str] = None,
    description: Optional[str] = None,
    default: Any = Undefined,
    examples: Optional[Sequence[Any]] = None,
    deprecated: Deprecated = False,
    extra: Mapping[str, Any] = None,
    override: bool = False,
) -> Schema:
    ...


@overload
def schema(
    *,
    title: Optional[str] = None,
    description: Optional[str] = None,
    default: Any = Undefined,
    examples: Optional[Sequence[Any]] = None,
    deprecated: Deprecated = False,
    min: Optional[Number] = None,
    max: Optional[Number] = None,
    exc_min: Optional[Number] = None,
    exc_max: Optional[Number] = None,
    mult_of: Optional[Number] = None,
    extra: Mapping[str, Any] = None,
    override: bool = False,
) -> Schema:
    ...


@overload
def schema(
    *,
    title: Optional[str] = None,
    description: Optional[str] = None,
    default: Any = Undefined,
    examples: Optional[Sequence[Any]] = None,
    deprecated: Deprecated = False,
    format: Optional[str] = None,
    min_len: Optional[int] = None,
    max_len: Optional[int] = None,
    pattern: Optional[Union[str, Pattern]] = None,
    extra: Mapping[str, Any] = None,
    override: bool = False,
) -> Schema:
    ...


@overload
def schema(
    *,
    title: Optional[str] = None,
    description: Optional[str] = None,
    default: Any = Undefined,
    examples: Optional[Sequence[Any]] = None,
    deprecated: Deprecated = False,
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique: Optional[bool] = None,
    extra: Mapping[str, Any] = None,
    override: bool = False,
) -> Schema:
    ...


@overload
def schema(
    *,
    title: Optional[str] = None,
    description: Optional[str] = None,
    default: Any = Undefined,
    examples: Optional[Sequence[Any]] = None,
    deprecated: Deprecated = False,
    min_props: Optional[int] = None,
    max_props: Optional[int] = None,
    extra: Mapping[str, Any] = None,
    override: bool = False,
) -> Schema:
    ...


def schema(override=False, **kwargs) -> Schema:
    annotations_kwargs = {k: v for k, v in kwargs.items() if k in _annotations_fields}
    constraints_kwargs = {
        k: v for k, v in kwargs.items() if k not in _annotations_fields
    }
    annotations = Annotations(**annotations_kwargs) if annotations_kwargs else None

    constraints = None
    if constraints_kwargs:
        for cls in (
            NumberConstraints,
            StringConstraints,
            ArrayConstraints,
            ObjectConstraints,
        ):
            with suppress(TypeError):
                constraints = cls(**constraints_kwargs)
                break
        else:
            raise TypeError("Invalid schema")

    return Schema(annotations, constraints, override)


def _default_schema(tp: AnyType) -> Optional[Schema]:
    return None


_schemas: Dict[Any, Schema] = {}


def get_schema(tp: AnyType) -> Optional[Schema]:
    tp = replace_builtins(tp)
    return _schemas[tp] if contains(_schemas, tp) else _default_schema(tp)


@merge_opts
def merge_schema(default: Schema, override: Schema) -> Schema:
    if override.override:
        return override
    return Schema(
        merge_annotations(default.annotations, override.annotations),
        merge_constraints(default.constraints, override.constraints),
        override.override,
    )



__all__ = ["fields", "fields_set", "set_fields", "unset_fields", "with_fields_set"]
from dataclasses import (  # type: ignore
    Field,
    _FIELD,
    _FIELDS,
    _FIELD_CLASSVAR,
    _FIELD_INITVAR,
    fields as fields_,
    is_dataclass,
)
from functools import wraps
from inspect import signature
from typing import (
    AbstractSet,
    Any,
    Collection,
    Set,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

from apischema.utils import PREFIX

FIELDS_SET_ATTR = f"{PREFIX}fields_set"


def _check_dataclass(obj):
    if not is_dataclass(obj):
        raise ValueError("not a dataclass")


Cls = TypeVar("Cls", bound=Type)

_ALREADY_SET = f"{PREFIX}already_set"


def with_fields_set(cls: Cls) -> Cls:
    from apischema.metadata.keys import DEFAULT_AS_SET

    init_fields = set()
    post_init_fields = set()
    if is_dataclass(cls):
        for field in getattr(cls, _FIELDS).values():
            assert isinstance(field, Field)
            if field._field_type == _FIELD_INITVAR:  # type: ignore
                init_fields.add(field.name)
            if field._field_type == _FIELD and not field.init:  # type: ignore
                post_init_fields.add(field.name)
            if field.metadata.get(DEFAULT_AS_SET):
                post_init_fields.add(field.name)
    params = list(signature(cls.__init__).parameters)[1:]
    old_new = cls.__new__
    old_init = cls.__init__
    old_setattr = cls.__setattr__

    def new_new(*args, **kwargs):
        if old_new is object.__new__:
            obj = object.__new__(args[0])
        else:
            obj = old_new(*args, **kwargs)
        # Initialize FIELD_SET_ATTR in order to prevent inherited class which override
        # __init__ to raise in __setattr__
        obj.__dict__[FIELDS_SET_ATTR] = set()
        return obj

    def new_init(self, *args, **kwargs):
        prev_fields_set = self.__dict__.get(FIELDS_SET_ATTR, set()).copy()
        self.__dict__[FIELDS_SET_ATTR] = set()
        try:
            old_init(self, *args, **kwargs)
        except TypeError as err:
            if str(err) == no_dataclass_init_error:
                raise RuntimeError(dataclass_before_error) from None
            else:
                raise
        arg_fields = {*params[: len(args)], *kwargs} - init_fields
        self.__dict__[FIELDS_SET_ATTR] = prev_fields_set | arg_fields | post_init_fields

    def new_setattr(self, attr, value):
        try:
            self.__dict__[FIELDS_SET_ATTR].add(attr)
        except KeyError:
            raise RuntimeError(dataclass_before_error) from None
        old_setattr(self, attr, value)

    for attr, old, new in [
        ("__new__", old_new, new_new),
        ("__init__", old_init, new_init),
        ("__setattr__", old_setattr, new_setattr),
    ]:
        if hasattr(old, _ALREADY_SET):
            continue
        setattr(new, _ALREADY_SET, True)
        setattr(cls, attr, wraps(old)(new))  # type: ignore

    return cls


no_dataclass_init_error = (
    "object.__init__() takes exactly one argument (the instance to initialize)"
)
dataclass_before_error = (
    f"{with_fields_set.__name__} must be put before dataclass decorator"
)


T = TypeVar("T")


def _all_fields(obj: Any) -> AbstractSet[str]:
    _check_dataclass(obj)
    return {
        f.name
        for f in fields_(obj)
        if f._field_type is not _FIELD_INITVAR  # type: ignore
    }


def _field_names(obj: Any, fields: Collection) -> AbstractSet[str]:
    all_fields = _all_fields(obj)
    result: Set[str] = set()
    for f in fields:
        if isinstance(f, Field):
            f = f.name
        if not isinstance(f, str):
            raise ValueError("Fields must be dataclass Field or str")
        if f not in all_fields:
            raise ValueError(f"Wrong field {f}")
        result.add(f)
    return result


def set_fields(obj: T, *fields: Any, overwrite=False) -> T:
    _fields = _field_names(obj, fields)
    if overwrite:
        obj.__dict__[FIELDS_SET_ATTR] = _fields
    else:
        try:
            obj.__dict__[FIELDS_SET_ATTR].update(_fields)
        except KeyError:
            # with_fields_set is not use, so all fields are set
            pass
    return obj


def unset_fields(obj: T, *fields: Any) -> T:
    _fields = _field_names(obj, fields)
    obj.__dict__[FIELDS_SET_ATTR] = fields_set(obj) - _fields
    return obj


def fields_set(obj: Any) -> AbstractSet[str]:
    try:
        return getattr(obj, FIELDS_SET_ATTR)
    except AttributeError:
        return _all_fields(obj)


class FieldGetter:
    def __init__(self, obj):
        _check_dataclass(obj)
        self.fields = {
            name: f
            for name, f in getattr(obj, _FIELDS).items()
            if not f._field_type == _FIELD_CLASSVAR
        }

    def __getattribute__(self, name: str) -> Field:
        try:
            return super().__getattribute__("fields")[name]
        except KeyError:
            raise AttributeError(name)


@overload
def fields(obj: Type[T]) -> T:
    ...


@overload
def fields(obj: T) -> T:
    ...


# Overload because of Mypy issue
# https://github.com/python/mypy/issues/9003#issuecomment-667418520
def fields(obj: Union[Type[T], T]) -> T:
    return cast(T, FieldGetter(obj))


class FieldIsSet(FieldGetter):
    def __init__(self, obj):
        super().__init__(obj)
        self.obj = obj

    def __getattribute__(self, name: str) -> bool:  # type: ignore
        field = super().__getattribute__(name)
        return field.name in fields_set(object.__getattribute__(self, "obj"))


def is_set(obj: T) -> T:
    return cast(T, FieldIsSet(obj))



from dataclasses import dataclass
from typing import Any, Callable, ClassVar, Dict, Mapping, Optional

from apischema.conversions import Conversion, LazyConversion
from apischema.json_schema.types import JsonSchema

RefFactory = Callable[[str], str]


def ref_prefix(prefix: str) -> RefFactory:
    if not prefix.endswith("/"):
        prefix += "/"
    return lambda ref: prefix + ref


def isolate_ref(schema: Dict[str, Any]):
    if "$ref" in schema and len(schema) > 1:
        schema.setdefault("allOf", []).append({"$ref": schema.pop("$ref")})


def to_json_schema_7(schema: JsonSchema) -> Mapping[str, Any]:
    result = schema.copy()
    isolate_ref(result)
    if "$defs" in result:
        result["definitions"] = {**result.pop("$defs"), **result.get("definitions", {})}
    if "dependentRequired" in result:
        result["dependencies"] = {
            **result.pop("dependentRequired"),
            **result.get("dependencies", {}),
        }
    return result


def to_open_api_3_0(schema: JsonSchema) -> Mapping[str, Any]:
    result = schema.copy()
    for key in ("dependentRequired", "unevaluatedProperties", "$defs"):
        result.pop(key, ...)
    isolate_ref(result)
    if "null" in result.get("type", ()):
        result.setdefault("nullable", True)
        if result["type"] == "null":
            result.pop("type")
        else:
            types = [t for t in result["type"] if t != "null"]
            result["type"] = types if len(types) > 1 else types[0]
    if {"type": "null"} in result.get("anyOf", ()):
        result.setdefault("nullable", True)
        result["anyOf"] = [a for a in result["anyOf"] if a != {"type": "null"}]
    if "examples" in result:
        result.setdefault("example", result.pop("examples")[0])
    return result


@dataclass
class JsonSchemaVersion:
    schema: Optional[str] = None
    ref_prefix: str = ""
    serialization: Optional[Callable] = None
    all_refs: bool = True

    @property
    def conversion(self) -> Optional[Conversion]:
        if self.serialization:
            # Recursive conversion pattern
            tmp = None
            conversion = Conversion(
                self.serialization, sub_conversions=LazyConversion(lambda: tmp)
            )
            tmp = conversion
            return conversion
        else:
            return None

    @property
    def ref_factory(self) -> RefFactory:
        return ref_prefix(self.ref_prefix)

    DRAFT_2019_09: ClassVar["JsonSchemaVersion"]
    DRAFT_7: ClassVar["JsonSchemaVersion"]
    OPEN_API_3_0: ClassVar["JsonSchemaVersion"]


JsonSchemaVersion.DRAFT_2019_09 = JsonSchemaVersion(
    "http://json-schema.org/draft/2019-09/schema#", "#/$defs/", None, False
)
JsonSchemaVersion.DRAFT_7 = JsonSchemaVersion(
    "http://json-schema.org/draft-07/schema#", "#/definitions/", to_json_schema_7, False
)
JsonSchemaVersion.OPEN_API_3_0 = JsonSchemaVersion(
    None, "#/components/schemas/", to_open_api_3_0, True
)



__all__ = [
    "Conversion",
    "LazyConversion",
    "as_str",
    "dataclass_input_wrapper",
    "dataclass_model",
    "deserializer",
    "identity",
    "inherited_deserializer",
    "reset_deserializers",
    "reset_serializer",
    "serializer",
]

from .conversions import Conversion, LazyConversion
from .converters import (
    as_str,
    deserializer,
    inherited_deserializer,
    reset_deserializers,
    reset_serializer,
    serializer,
)
from .dataclass_models import dataclass_model
from .utils import identity
from .wrappers import dataclass_input_wrapper



__all__ = ["Tagged", "TaggedUnion", "get_tagged"]
from dataclasses import dataclass, field
from typing import (
    Any,
    ClassVar,
    Generic,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

from apischema.aliases import alias
from apischema.conversions.conversions import ConvOrFunc
from apischema.json_schema.schema import Schema
from apischema.metadata import conversion
from apischema.types import Metadata, MetadataImplem
from apischema.typing import get_type_hints
from apischema.utils import PREFIX, Undefined, UndefinedType, get_args2, get_origin2

TAGS_ATTR = f"{PREFIX}tags"

T = TypeVar("T", bound="TaggedUnion")
V = TypeVar("V")


class Tag(str, Generic[T, V]):
    def __new__(cls, tag: str, type: Type[T]):
        return super().__new__(cls, tag)

    def __init__(self, tag: str, type: Type[T]):
        super().__init__()
        self.type = type

    def __call__(self, value: V) -> T:
        return self.type(**{self: value})  # type: ignore


@dataclass(frozen=True)
class Tagged(Generic[V]):
    alias: Optional[str] = None
    schema: Optional[Schema] = None
    deserialization: Optional[ConvOrFunc] = None
    serialization: Optional[ConvOrFunc] = None

    @property
    def metadata(self) -> Metadata:
        metadata: Metadata = MetadataImplem()
        if self.alias is not None:
            metadata |= alias(self.alias)
        if self.schema is not None:
            metadata |= self.schema
        if self.deserialization is not None or self.serialization is not None:
            metadata |= conversion(self.deserialization, self.serialization)
        return metadata

    @overload
    def __get__(self, instance: None, owner: Type[T]) -> Tag[T, V]:
        ...

    @overload
    def __get__(self, instance: Any, owner) -> Union[V, UndefinedType]:
        ...

    def __get__(self, instance, owner):
        raise NotImplementedError


class TaggedUnion:
    def __init__(self, **kwargs):
        if len(kwargs) != 1:
            raise ValueError(
                f"tagged union must have one and only one tag set,"
                f" found {list(kwargs)}"
            )
        tags = getattr(self, TAGS_ATTR)
        for tag in tags:
            setattr(self, tag, Undefined)
        for tag, value in kwargs.items():
            if tag not in tags:
                raise TypeError(f"{type(self)} has no tag {tag}")
            setattr(self, tag, value)

    def __repr__(self):
        tag, value = get_tagged(self)
        return f"{type(self).__name__}({tag}={value!r})"

    def __init_subclass__(cls, **kwargs):
        tags = set(getattr(cls, TAGS_ATTR, ()))
        types = get_type_hints(cls, include_extras=True)
        for tag, tp in types.items():
            if get_origin2(tp) == Tagged:
                if tag in tags:
                    raise TypeError(f"Cannot redefine tag {tag} in {cls}")
                tagged = cls.__dict__.get(tag, Tagged())
                setattr(cls, tag, field(default=Undefined, metadata=tagged.metadata))
                cls.__annotations__[tag] = Union[
                    get_args2(types[tag])[0], UndefinedType
                ]
                tags.add(tag)
            elif tag not in tags and get_origin2(tp) != ClassVar:
                cls.__annotations__[tag] = ClassVar[tp]
        setattr(cls, TAGS_ATTR, tags)
        dataclass(init=False, repr=False)(cls)
        for tag in tags:
            setattr(cls, tag, Tag(tag, cls))


def get_tagged(tagged_union: TaggedUnion) -> Tuple[str, Any]:
    defined = {
        tag: getattr(tagged_union, tag)
        for tag in getattr(tagged_union, TAGS_ATTR)
        if getattr(tagged_union, tag) is not Undefined
    }
    return next(iter(defined.items()))



from functools import wraps
from typing import Any, Callable, Dict, Type, TypeVar, Union

from apischema.json_schema.types import JsonType
from apischema.types import NoneType
from apischema.validation.errors import ValidationError

T = TypeVar("T")

Coercer = Callable[[Type[T], Any], T]


def no_coercion(expected: Type[T], data: Any) -> T:
    if not isinstance(data, expected):
        if expected is float and isinstance(data, int):
            return float(data)  # type: ignore
        msg = (
            f"expected type {JsonType.from_type(expected)},"
            f" found {JsonType.from_type(type(data))}"
        )
        raise ValidationError([msg])
    return data


_bool_pairs = (
    ("0", "1"),
    ("f", "t"),
    ("n", "y"),
    ("no", "yes"),
    ("false", "true"),
    ("off", "on"),
    ("ko", "ok"),
)
STR_TO_BOOL: Dict[str, bool] = {}
for false, true in _bool_pairs:
    for s, value in ((false, False), (true, True)):
        STR_TO_BOOL[s.lower()] = value
STR_NONE_VALUES = {""}


def coercion_error(cls: Type, data) -> ValidationError:
    msg = (
        f"cannot coerce {JsonType.from_type(cls)}"
        f" from {JsonType.from_type(type(data))}"
    )
    return ValidationError([msg])


def coerce(cls: Type[T], data: Any) -> T:
    try:
        if isinstance(data, cls):
            return data
        if data is None and cls is not NoneType:
            raise ValueError
        if cls is bool and isinstance(data, str):
            return STR_TO_BOOL[data.lower()]  # type: ignore
        if cls is NoneType and data in STR_NONE_VALUES:
            return None  # type: ignore
        if cls is list and isinstance(data, str):
            raise ValueError
        return cls(data)  # type: ignore
    except (ValueError, TypeError, KeyError):
        raise coercion_error(cls, data) from None


_coercer: Coercer = coerce

Coercion = Union[bool, Coercer]


def wrap_coercer(coercer: Coercer) -> Coercer:
    @wraps(coercer)
    def wrapper(cls, data):
        try:
            result = coercer(cls, data)
        except AssertionError:
            raise
        except Exception:
            raise coercion_error(cls, data)
        if not isinstance(result, cls):
            raise coercion_error(cls, data)
        return result

    return wrapper


def get_coercer(coercion: Coercion) -> Coercer:
    if callable(coercion):
        return wrap_coercer(coercion)
    elif coercion:
        return _coercer
    else:
        return no_coercion



__all__ = [
    "alias",
    "conversion",
    "default_as_set",
    "default_fallback",
    "init_var",
    "merged",
    "post_init",
    "properties",
    "required",
    "skip",
    "validators",
]

from apischema.aliases import alias
from .implem import (
    conversion,
    default_as_set,
    default_fallback,
    init_var,
    merged,
    post_init,
    properties,
    required,
    skip,
    validators,
)



from dataclasses import Field, is_dataclass
from enum import Enum
from functools import lru_cache
from types import MappingProxyType
from typing import (
    Any,
    Collection,
    Generic,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from apischema.dataclass_utils import dataclass_types_and_fields
from apischema.types import (
    AnyType,
    COLLECTION_TYPES,
    MAPPING_TYPES,
    OrderedDict,
    PRIMITIVE_TYPES,
)
from apischema.typing import (
    _LiteralMeta,
    _TypedDictMeta,
    get_args,
    get_origin,
    get_type_hints2,
)
from apischema.utils import is_type_var

try:
    from apischema.typing import Annotated, Literal
except ImportError:
    Annotated, Literal = ..., ...  # type: ignore

TUPLE_TYPE = get_origin(Tuple[Any])


@lru_cache()
def type_hints_cache(obj) -> Mapping[str, AnyType]:
    # Use immutable return because of cache
    return MappingProxyType(get_type_hints2(obj))


class Unsupported(TypeError):
    def __init__(self, cls: Type):
        self.cls = cls


Return = TypeVar("Return", covariant=True)


class Visitor(Generic[Return]):
    def __init__(self):
        self._generic: Optional[AnyType] = None

    def annotated(self, tp: AnyType, annotations: Sequence[Any]) -> Return:
        return self.visit(tp)

    def any(self) -> Return:
        raise NotImplementedError

    def collection(self, cls: Type[Collection], value_type: AnyType) -> Return:
        raise NotImplementedError

    def dataclass(
        self,
        cls: Type,
        types: Mapping[str, AnyType],
        fields: Sequence[Field],
        init_vars: Sequence[Field],
    ) -> Return:
        raise NotImplementedError

    def enum(self, cls: Type[Enum]) -> Return:
        raise NotImplementedError

    def generic(self, tp: AnyType) -> Return:
        _generic = self._generic
        self._generic = tp
        try:
            return self._visit(get_origin(tp))
        finally:
            self._generic = _generic

    def literal(self, values: Sequence[Any]) -> Return:
        raise NotImplementedError

    def mapping(
        self, cls: Type[Mapping], key_type: AnyType, value_type: AnyType
    ) -> Return:
        raise NotImplementedError

    def named_tuple(
        self,
        cls: Type[Tuple],
        types: Mapping[str, AnyType],
        defaults: Mapping[str, Any],
    ) -> Return:
        raise NotImplementedError

    def new_type(self, tp: AnyType, super_type: AnyType) -> Return:
        return self.visit(super_type)

    def primitive(self, cls: Type) -> Return:
        raise NotImplementedError

    def subprimitive(self, cls: Type, superclass: Type) -> Return:
        return self.primitive(superclass)

    def tuple(self, types: Sequence[AnyType]) -> Return:
        raise NotImplementedError

    def typed_dict(self, cls: Type, keys: Mapping[str, AnyType], total: bool) -> Return:
        raise NotImplementedError

    def _union_result(self, results: Iterable[Return]) -> Return:
        raise NotImplementedError

    def union(self, alternatives: Sequence[AnyType]) -> Return:
        return self._union_result(map(self.visit, alternatives))

    def unsupported(self, tp: AnyType) -> Return:
        raise Unsupported(tp)

    def _visit_generic(self, tp: AnyType) -> Return:
        origin, args = get_origin(tp), get_args(tp)
        assert origin is not None
        if origin is Annotated:
            return self.annotated(args[0], args[1:])
        if origin is Union:
            return self.union(args)
        if origin is TUPLE_TYPE:
            if len(args) < 2 or args[1] is not ...:
                return self.tuple(args)
        if origin in COLLECTION_TYPES:
            return self.collection(origin, args[0])
        if origin in MAPPING_TYPES:
            return self.mapping(origin, args[0], args[1])
        if origin is Literal:  # pragma: no cover py37+
            return self.literal(args)
        return self.generic(tp)

    def _visit(self, tp: AnyType) -> Return:
        if tp in PRIMITIVE_TYPES:
            return self.primitive(tp)
        if is_dataclass(tp):
            return self.dataclass(
                tp, *dataclass_types_and_fields(self._generic or tp)  # type: ignore
            )
        if hasattr(tp, "__supertype__"):
            return self.new_type(tp, tp.__supertype__)
        if tp is Any:
            return self.any()
        if tp in COLLECTION_TYPES:
            return self.collection(tp, Any)
        if tp in MAPPING_TYPES:
            return self.mapping(tp, Any, Any)
        try:
            issubclass(tp, object)
        except TypeError:
            pass
        else:
            if issubclass(tp, Enum):
                return self.enum(tp)
            for primitive in PRIMITIVE_TYPES:
                if issubclass(tp, primitive):
                    return self.subprimitive(tp, primitive)
            # NamedTuple
            if issubclass(tp, tuple) and hasattr(tp, "_fields"):
                if hasattr(tp, "__annotations__"):
                    types = type_hints_cache(self._generic or tp)
                elif hasattr(tp, "__field_types"):  # pragma: no cover
                    types = tp._field_types  # type: ignore
                else:  # pragma: no cover
                    types = OrderedDict((f, Any) for f in tp._fields)  # type: ignore
                return self.named_tuple(tp, types, tp._field_defaults)  # type: ignore
        if isinstance(tp, _LiteralMeta):  # pragma: no cover py36
            return self.literal(tp.__values__)  # type: ignore
        # cannot use issubclass(..., TypedDict)
        if isinstance(tp, _TypedDictMeta):
            total = tp.__total__  # type: ignore
            assert isinstance(tp, type)
            return self.typed_dict(tp, type_hints_cache(self._generic or tp), total)
        return self.unsupported(tp)

    def visit(self, tp: AnyType) -> Return:
        if get_origin(tp) is not None:
            return self._visit_generic(tp)
        if is_type_var(tp):
            if tp.__constraints__:
                return self.visit(Union[tp.__constraints__])
            else:
                return self.visit(Any)
        else:
            _generic = self._generic
            self._generic = None
            try:
                return self._visit(tp)
            finally:
                self._generic = _generic



from dataclasses import MISSING, field as field_, make_dataclass
from inspect import Parameter, iscoroutinefunction, signature
from typing import Any, Callable, Mapping, Tuple, Type

from apischema.metadata import properties
from apischema.metadata.keys import (
    ALIAS_METADATA,
    ALIAS_NO_OVERRIDE_METADATA,
    CONVERSIONS_METADATA,
    DEFAULT_AS_SET,
    DEFAULT_FALLBACK_METADATA,
    REQUIRED_METADATA,
    get_annotated_metadata,
)
from apischema.typing import get_type_hints
from apischema.utils import to_camel_case

FIELDS_METADATA = {
    ALIAS_METADATA,
    ALIAS_NO_OVERRIDE_METADATA,
    CONVERSIONS_METADATA,
    DEFAULT_AS_SET,
    DEFAULT_FALLBACK_METADATA,
    REQUIRED_METADATA,
}


def dataclass_input_wrapper(
    func: Callable, parameters_metadata: Mapping[str, Mapping] = None
) -> Tuple[Callable, Type]:
    parameters_metadata = parameters_metadata or {}
    types = get_type_hints(func, include_extras=True)
    fields = []
    params, kwargs_param = [], None
    for param_name, param in signature(func).parameters.items():
        if param.kind is Parameter.POSITIONAL_ONLY:
            raise TypeError("Positional only parameters are not supported")
        field_type = types.get(param_name, Any)
        if param.kind in {Parameter.POSITIONAL_OR_KEYWORD, Parameter.KEYWORD_ONLY}:
            metadata = get_annotated_metadata(field_type)
            if param_name in parameters_metadata:
                metadata = {**metadata, **parameters_metadata[param_name]}
            metadata = {k: v for k, v in metadata.items() if k in FIELDS_METADATA}
            default = MISSING if param.default is Parameter.empty else param.default
            field = field_(default=default, metadata=metadata)
            fields.append((param_name, field_type, field))
            params.append(param_name)
        if param.kind == Parameter.VAR_KEYWORD:
            field = field_(default_factory=dict, metadata=properties)
            fields.append((param_name, Mapping[str, field_type], field))  # type: ignore
            kwargs_param = param_name

    input_cls = make_dataclass(to_camel_case(func.__name__), fields)

    def wrapper(input):
        kwargs = {name: getattr(input, name) for name in params}
        if kwargs_param:
            kwargs.update(getattr(input, kwargs_param))
        return func(**kwargs)

    if iscoroutinefunction(func):
        wrapped = wrapper

        async def wrapper(input):
            return await wrapped(input)

    wrapper.__annotations__["input"] = input_cls
    if "return" in func.__annotations__:
        wrapper.__annotations__["return"] = func.__annotations__["return"]
    return wrapper, input_cls



import sys
from dataclasses import (  # type: ignore
    Field,
    InitVar,
    MISSING,
    _FIELDS,
    _FIELD_CLASSVAR,
    is_dataclass,
    make_dataclass,
)
from functools import lru_cache
from itertools import chain
from types import MappingProxyType
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    Mapping,
    Optional,
    Sequence,
    Set,
    TYPE_CHECKING,
    Tuple,
    Type,
)

from apischema.dependent_required import (
    DEPENDENT_REQUIRED_ATTR,
    DependentRequired,
    Requirements,
)
from apischema.metadata.implem import ConversionMetadata
from apischema.metadata.keys import ALIAS_METADATA, CONVERSIONS_METADATA
from apischema.types import AnyType
from apischema.typing import get_type_hints, get_type_hints2
from apischema.utils import (
    OperationKind,
    PREFIX,
    get_origin2,
    get_origin_or_type,
    has_type_vars,
)

if TYPE_CHECKING:
    from apischema.conversions.conversions import Conversions

try:
    from apischema.typing import Annotated
except ImportError:
    Annotated = ...  # type: ignore

if sys.version_info <= (3, 7):  # pragma: no cover
    is_dataclass_ = is_dataclass

    def is_dataclass(obj) -> bool:
        return is_dataclass_(obj) and getattr(obj, "__origin__", None) is None


@lru_cache()
def dataclass_types_and_fields(
    tp: AnyType,
) -> Tuple[Mapping[str, AnyType], Sequence[Field], Sequence[Field]]:
    from apischema.metadata.keys import INIT_VAR_METADATA

    cls = get_origin_or_type(tp)
    assert is_dataclass(cls)
    types = get_type_hints2(tp)
    fields, init_fields = [], []
    for field in getattr(cls, _FIELDS).values():
        assert isinstance(field, Field)
        if field._field_type == _FIELD_CLASSVAR:  # type: ignore
            continue
        field_type = types[field.name]
        if isinstance(field_type, InitVar):
            types[field.name] = field_type.type  # type: ignore
            init_fields.append(field)
        elif field_type is InitVar:
            metadata = getattr(cls, _FIELDS)[field.name].metadata
            if INIT_VAR_METADATA not in metadata:
                raise TypeError("Before 3.8, InitVar requires init_var metadata")
            init_field = (PREFIX, metadata[INIT_VAR_METADATA], ...)
            tmp_cls = make_dataclass("Tmp", [init_field], bases=(cls,))  # type: ignore
            types[field.name] = get_type_hints(tmp_cls, include_extras=True)[PREFIX]
            if has_type_vars(types[field.name]):
                raise TypeError("Generic InitVar are not supported before 3.8")
            init_fields.append(field)
        else:
            fields.append(field)
    # Use immutable return because of cache
    return MappingProxyType(types), tuple(fields), tuple(init_fields)


def get_fields(
    fields: Sequence[Field], init_vars: Sequence[Field], operation: OperationKind
) -> Sequence[Field]:
    from apischema.metadata.keys import SKIP_METADATA

    fields_by_operation = {
        OperationKind.DESERIALIZATION: chain((f for f in fields if f.init), init_vars),
        operation.SERIALIZATION: fields,
    }[operation]
    return [f for f in fields_by_operation if SKIP_METADATA not in f.metadata]


def has_default(field: Field) -> bool:
    return field.default is not MISSING or field.default_factory is not MISSING  # type: ignore # noqa: E501


def is_required(field: Field) -> bool:
    from apischema.metadata.keys import REQUIRED_METADATA

    return REQUIRED_METADATA in field.metadata or not has_default(field)


def get_default(field: Field) -> Any:
    if field.default_factory is not MISSING:  # type: ignore
        return field.default_factory()  # type: ignore
    if field.default is not MISSING:
        return field.default
    raise NotImplementedError


def get_alias(field: Field) -> str:
    return field.metadata.get(ALIAS_METADATA, field.name)


def get_requirements(
    cls: Type,
    method: Callable[[DependentRequired], Requirements],
    operation: OperationKind,
) -> Requirements:
    assert is_dataclass(cls)
    _, fields, init_vars = dataclass_types_and_fields(cls)  # type: ignore
    all_dependent_required: Collection["DependentRequired"] = getattr(
        cls, DEPENDENT_REQUIRED_ATTR, ()
    )
    requirements: Dict[Field, Set[Field]] = {}
    fields_by_operation = get_fields(fields, init_vars, operation)
    for dep_req in all_dependent_required:
        for field, required in method(dep_req).items():  # noqa: F402
            requirements.setdefault(field, set()).update(
                req for req in required if req in fields_by_operation
            )
    return requirements


def check_merged_class(merged_cls: AnyType) -> Type:
    origin = get_origin2(merged_cls)
    if origin is None:
        origin = merged_cls
    if not is_dataclass(origin):
        raise TypeError("Merged field must have dataclass type")
    return origin


def get_field_conversions(
    field: Field, operation: OperationKind
) -> Optional["Conversions"]:

    if CONVERSIONS_METADATA not in field.metadata:
        return None
    else:
        conversions: ConversionMetadata = field.metadata[CONVERSIONS_METADATA]
        return {
            OperationKind.DESERIALIZATION: conversions.deserialization,
            OperationKind.SERIALIZATION: conversions.serialization,
        }[operation]



import warnings
from contextlib import contextmanager
from dataclasses import replace
from types import new_class
from typing import (
    Any,
    Collection,
    Generic,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from apischema.conversions.conversions import (
    Conversions,
    ResolvedConversion,
    ResolvedConversions,
    handle_container_conversions,
    is_identity,
    resolve_conversions,
)
from apischema.conversions.converters import _deserializers, _serializers
from apischema.conversions.dataclass_models import DataclassModel
from apischema.conversions.utils import (
    INVALID_CONVERSION_TYPES,
)
from apischema.skip import filter_skipped
from apischema.types import AnyType, COLLECTION_TYPES, MAPPING_TYPES
from apischema.typing import generic_mro, get_args, get_origin
from apischema.utils import (
    OperationKind,
    PREFIX,
    Undefined,
    UndefinedType,
    get_args2,
    get_origin_or_type,
    is_type_var,
    substitute_type_vars,
)
from apischema.visitor import Return, Visitor

try:
    from apischema.typing import Annotated
except ImportError:
    Annotated = ...  # type: ignore

Deserialization = Sequence[ResolvedConversion]
Serialization = ResolvedConversion
Conv = TypeVar("Conv", Deserialization, Serialization)

SELF_CONVERSION_ATTR = f"{PREFIX}self_conversion"


class ConversionsVisitor(Visitor[Return], Generic[Conv, Return]):
    operation: OperationKind
    _dynamic_conversion_resolver: Type["DynamicConversionResolver"]

    def __init__(self):
        super().__init__()
        self._conversions: ResolvedConversions = ()

    @staticmethod
    def _get_conversions(
        tp: Type, conversions: ResolvedConversions
    ) -> Union[Conv, None, UndefinedType]:
        raise NotImplementedError

    @staticmethod
    def _default_conversions(tp: Type) -> Optional[Conversions]:
        raise NotImplementedError

    @classmethod
    def get_conversions(
        cls, tp: Type, conversions: ResolvedConversions
    ) -> Tuple[Optional[Conv], bool]:
        result, dynamic = None, False
        if conversions is not None:
            result = cls._get_conversions(tp, conversions)
        if result is not None:
            dynamic = True
        else:
            default_conv = cls._default_conversions(tp)
            if default_conv:
                result = cls._get_conversions(tp, resolve_conversions(default_conv))
        return (result if result is not Undefined else None), dynamic  # type: ignore

    def _apply_dynamic_conversions(self, tp: AnyType) -> Optional[AnyType]:
        if not self._conversions:
            return None
        else:
            return self._dynamic_conversion_resolver().visit_with_conversions(
                tp, self._conversions
            )

    def visit_conversion(self, tp: AnyType, conversion: Conv, dynamic: bool) -> Return:
        raise NotImplementedError

    def visit_not_conversion(self, tp: AnyType) -> Return:
        return super().visit(tp)

    @contextmanager
    def _replace_conversions(self, conversions: Optional[Conversions]):
        conversions_save = self._conversions
        self._conversions = resolve_conversions(conversions)
        try:
            yield
        finally:
            self._conversions = conversions_save

    def visit(self, tp: AnyType) -> Return:
        origin = get_origin(tp) or tp
        if origin in INVALID_CONVERSION_TYPES or not isinstance(origin, type):
            return self.visit_not_conversion(tp)
        conversion, dynamic = self.get_conversions(origin, self._conversions)
        if conversion is None:
            with self._replace_conversions(
                handle_container_conversions(tp, None, self._conversions, dynamic)
            ):
                return self.visit_not_conversion(tp)
        else:
            return self.visit_conversion(tp, conversion, dynamic)

    @staticmethod
    def _update_generic_args(tp: AnyType, conversion: ResolvedConversion) -> AnyType:
        raise NotImplementedError

    def visit_with_conversions(
        self, tp: AnyType, conversions: Optional[Conversions]
    ) -> Return:
        with self._replace_conversions(conversions):
            return self.visit(tp)


class DeserializationVisitor(ConversionsVisitor[Deserialization, Return]):
    operation = OperationKind.DESERIALIZATION

    @staticmethod
    def _get_conversions(
        tp: Type, conversions: ResolvedConversions
    ) -> Union[Deserialization, None, UndefinedType]:
        origin = get_origin_or_type(tp)
        result = [
            conv
            for conv in conversions
            if issubclass(get_origin_or_type(conv.target), origin)
        ]
        for i, conv in enumerate(result):
            if is_identity(conv):
                if len(result) == 1:
                    return Undefined
                else:
                    namespace = {
                        "__new__": lambda _, *args, **kwargs: origin(*args, **kwargs),
                        SELF_CONVERSION_ATTR: True,
                    }
                    wrapper = new_class(
                        f"{origin.__name__}SelfDeserializer",
                        (tp,),
                        exec_body=lambda ns: ns.update(namespace),
                    )
                    result[i] = ResolvedConversion(
                        replace(conv, source=wrapper, target=conv.target)
                    )
        return result or None

    _default_conversions = staticmethod(_deserializers.get)  # type: ignore

    @staticmethod
    def _update_generic_args(tp: AnyType, conversion: ResolvedConversion) -> AnyType:
        source = conversion.source
        if isinstance(source, DataclassModel):
            source = source.dataclass
        if is_type_var(conversion.target):
            return substitute_type_vars(source, {conversion.target: tp})
        elif get_origin(tp) is None:
            return source
        else:
            origin = get_origin_or_type(tp)
            if get_origin(conversion.target) == Annotated:
                target = get_args(conversion.target)[0]
            else:
                target = conversion.target
            if getattr(target, "__parameters__", ()):
                target = target[target.__parameters__]
            substitution = {}
            for base in generic_mro(target):
                base_origin = get_origin_or_type(base)
                if (
                    base_origin == origin
                    or (origin in MAPPING_TYPES and base_origin in MAPPING_TYPES)
                    or (
                        origin in COLLECTION_TYPES
                        and base_origin
                        in COLLECTION_TYPES.keys() | MAPPING_TYPES.keys()
                    )
                ):
                    for base_arg, arg in zip(get_args(base), get_args2(tp)):
                        if is_type_var(base_arg):
                            substitution[base_arg] = arg
                        elif base_arg != arg:
                            warnings.warn(
                                f"Generic conversion target {conversion.target} is"
                                f" incompatible with type {tp}"
                            )
                            return source
                break
            return substitute_type_vars(source, substitution)

    def visit_conversion(
        self, tp: AnyType, conversion: Deserialization, dynamic: bool
    ) -> Return:
        results = []
        for conv in conversion:
            results.append(
                self.visit_with_conversions(
                    self._update_generic_args(tp, conv),
                    handle_container_conversions(
                        conv.source, conv.sub_conversions, self._conversions, dynamic
                    ),
                )
            )
        return self._union_result(results)


class SerializationVisitor(ConversionsVisitor[Serialization, Return]):
    operation = OperationKind.SERIALIZATION

    @staticmethod
    def _get_conversions(
        tp: Type, conversions: ResolvedConversions
    ) -> Union[Serialization, None, UndefinedType]:
        origin = get_origin_or_type(tp)
        for conv in conversions:
            if issubclass(origin, get_origin_or_type(conv.source)):
                return Undefined if is_identity(conv) else conv
        else:
            return None

    @staticmethod
    def _default_conversions(tp: Type) -> Optional[Conversions]:
        for sub_cls in tp.__mro__:
            if sub_cls in _serializers:
                return _serializers[sub_cls]
        else:
            return None

    @staticmethod
    def _update_generic_args(tp: AnyType, conversion: ResolvedConversion) -> AnyType:
        target = conversion.target
        if isinstance(target, DataclassModel):
            target = target.dataclass
        if is_type_var(conversion.source):
            return substitute_type_vars(target, {conversion.source: tp})
        elif get_origin(tp) is None:
            return target
        else:
            source = conversion.source
            source_origin = get_origin_or_type(source)
            if getattr(source, "__parameters__", ()):
                source = source[source.__parameters__]
            substitution = {}
            for base in generic_mro(tp):
                base_origin = get_origin(base)
                if (
                    base_origin == source_origin
                    or (source_origin in MAPPING_TYPES and base_origin in MAPPING_TYPES)
                    or (
                        source_origin in COLLECTION_TYPES
                        and base_origin
                        in COLLECTION_TYPES.keys() | MAPPING_TYPES.keys()
                    )
                ):
                    for base_arg, source_arg in zip(get_args(base), get_args2(source)):
                        if is_type_var(source_arg):
                            substitution[source_arg] = base_arg
                        elif base_arg != source_arg:
                            warnings.warn(
                                f"Generic conversion source {conversion.source} is"
                                f" incompatible with type {tp}"
                            )
                            return target
                    break
            return substitute_type_vars(target, substitution)

    def visit_conversion(
        self, tp: AnyType, conversion: Serialization, dynamic: bool
    ) -> Return:
        return self.visit_with_conversions(
            self._update_generic_args(tp, conversion),
            handle_container_conversions(
                conversion.target,
                conversion.sub_conversions,
                self._conversions,
                dynamic,
            ),
        )


class DynamicConversionResolver(ConversionsVisitor[Conv, Optional[AnyType]]):
    def __init__(self):
        super().__init__()
        self._rec_guard: Set[Tuple[AnyType, ResolvedConversions]] = set()

    def annotated(self, tp: AnyType, annotations: Sequence[Any]) -> Optional[AnyType]:
        result = self.visit(tp)
        return Annotated[(result, *annotations)]

    def collection(
        self, cls: Type[Collection], value_type: AnyType
    ) -> Optional[AnyType]:
        value = self.visit(value_type)
        return None if value is None else Sequence[value]  # type: ignore

    def mapping(
        self, cls: Type[Mapping], key_type: AnyType, value_type: AnyType
    ) -> Optional[AnyType]:
        key = self.visit(key_type)
        value = self.visit(value_type)
        return None if key is None or value is None else Mapping[key, value]  # type: ignore # noqa: E501

    def visit_types(
        self, types: Iterable[AnyType], origin: AnyType
    ) -> Optional[AnyType]:
        modified = False
        types2 = []
        for tp in types:
            try:
                res = self.visit(tp)
                types2.append(res if res is not None else tp)
                modified = res is not None
            except Exception:
                types2.append(tp)
        return origin[tuple(types2)] if modified else None

    def tuple(self, types: Sequence[AnyType]) -> Optional[AnyType]:
        return self.visit_types(types, Tuple)

    def union(self, alternatives: Sequence[AnyType]) -> Optional[AnyType]:
        return self.visit_types(filter_skipped(alternatives, schema_only=True), Union)

    def _final_type(self, tp: AnyType, conversion: Conv) -> AnyType:
        raise NotImplementedError

    def visit(self, tp: AnyType) -> Optional[AnyType]:
        # if isinstance(tp, DataclassModel):
        #     return self.visit(tp.dataclass)
        origin = get_origin(tp) or tp
        if not self._conversions or (origin, self._conversions) in self._rec_guard:
            return None
        if origin in INVALID_CONVERSION_TYPES or not isinstance(origin, type):
            return self.visit_not_conversion(tp)
        conv, dynamic = self.get_conversions(origin, self._conversions)
        if not dynamic and (conv is None and get_origin(tp) is None):
            return None
        self._rec_guard.add((origin, self._conversions))
        try:
            if conv is not None:
                result = self.visit_conversion(tp, conv, dynamic)
            else:
                result = self.visit_not_conversion(tp)
            if result is not None:
                return result
            elif conv is not None:
                return self._final_type(tp, conv)
            else:
                return tp
        except Exception:
            return self._final_type(tp, conv) if conv is not None else tp
        finally:
            self._rec_guard.remove((origin, self._conversions))


class DynamicDeserializationResolver(DynamicConversionResolver, DeserializationVisitor):
    def _final_type(self, tp: AnyType, conversion: Deserialization) -> AnyType:
        return Union[tuple(self._update_generic_args(tp, conv) for conv in conversion)]

    def visit_conversion(
        self, tp: AnyType, conversion: Deserialization, dynamic: bool
    ) -> Optional[AnyType]:
        args = []
        modified = False
        for conv in conversion:
            source = self._update_generic_args(tp, conv)
            try:
                res = self.visit_with_conversions(
                    source,
                    handle_container_conversions(
                        conv.source, conv.sub_conversions, self._conversions, dynamic
                    ),
                )
                args.append(res if res is not None else source)
                modified = res is not None
            except Exception:
                args.append(source)
        return Union[tuple(args)] if modified else None


class DynamicSerializationResolver(DynamicConversionResolver, SerializationVisitor):
    def _final_type(self, tp: AnyType, conversion: Serialization) -> AnyType:
        return self._update_generic_args(tp, conversion)


DeserializationVisitor._dynamic_conversion_resolver = DynamicDeserializationResolver
SerializationVisitor._dynamic_conversion_resolver = DynamicSerializationResolver



from collections import ChainMap
from dataclasses import fields
from typing import Callable, TypeVar, overload

from apischema.types import MetadataImplem, Metadata

Aliaser = Callable[[str], str]
Cls = TypeVar("Cls")


@overload
def alias(alias_: str, *, override: bool = True) -> Metadata:
    ...


@overload
def alias(override: bool) -> Metadata:
    ...


@overload
def alias(aliaser: Aliaser) -> Callable[[Cls], Cls]:
    ...


def alias(arg=None, *, override: bool = True):  # type: ignore
    """Field alias or class aliaser

    :param alias_: alias of the field
    :param override: alias can be overridden by a class aliaser
    :param aliaser: compute alias for each (overridable) field of the class decorated
    """
    from apischema.metadata.keys import (
        ALIAS_METADATA,
        ALIAS_NO_OVERRIDE_METADATA,
        is_aggregate_field,
    )

    if callable(arg):

        def aliaser(cls: Cls) -> Cls:
            for field in fields(cls):
                if is_aggregate_field(field) or field.metadata.get(
                    ALIAS_NO_OVERRIDE_METADATA
                ):
                    continue
                alias = arg(field.metadata.get(ALIAS_METADATA, field.name))
                field.metadata = ChainMap({ALIAS_METADATA: alias}, field.metadata)
            return cls

        return aliaser
    metadata = {}
    if arg is not None:
        metadata[ALIAS_METADATA] = arg
    if not override:
        metadata[ALIAS_NO_OVERRIDE_METADATA] = True
    if not metadata:  # pragma: no cover
        raise ValueError("Alias must be called with arguments")
    return MetadataImplem(metadata)


def _global_aliaser(s: str) -> str:
    return s



from dataclasses import dataclass
from types import new_class
from typing import Callable, Optional, TYPE_CHECKING, Tuple, Type, Union

from apischema.conversions import Conversion
from apischema.conversions.utils import identity
from apischema.dataclasses import replace
from apischema.utils import PREFIX, cached_property

if TYPE_CHECKING:
    from apischema.deserialization.coercion import Coercion

Model = Union[Type, Callable[[], Type]]


def check_model(origin: Type, model: Type):
    if not isinstance(model, type):
        raise TypeError("Dataclass model must be a dataclass")
    if getattr(origin, "__parameters__", ()) != getattr(model, "__parameters__", ()):
        raise TypeError("Dataclass model must have the same generic parameters")


MODEL_ORIGIN_ATTR = f"{PREFIX}model_origin"


@dataclass(frozen=True)
class DataclassModel:
    origin: Type
    model: Model
    fields_only: bool

    @cached_property
    def dataclass(self) -> Type:
        origin = self.origin
        if isinstance(self.model, type):
            assert check_model(origin, self.model) is None
            model = self.model
        else:
            model = self.model()
            check_model(origin, model)
        namespace = {"__new__": lambda _, *args, **kwargs: origin(*args, **kwargs)}
        if not self.fields_only:
            namespace[MODEL_ORIGIN_ATTR] = origin
        return new_class(
            model.__name__, (model,), exec_body=lambda ns: ns.update(namespace)
        )


def dataclass_model(
    origin: Type,
    model: Model,
    *,
    fields_only: bool = False,
    additional_properties: Optional[bool] = None,
    coercion: Optional["Coercion"] = None,
    default_fallback: Optional[bool] = None,
    exclude_unset: Optional[bool] = None,
) -> Tuple[Conversion, Conversion]:
    if isinstance(model, type):
        check_model(origin, model)

    model_type = DataclassModel(origin, model, fields_only)
    conversion = Conversion(
        identity,
        additional_properties=additional_properties,
        coercion=coercion,
        default_fallback=default_fallback,
        exclude_unset=exclude_unset,
    )
    d_conv = replace(conversion, source=model_type, target=origin)
    s_conv = replace(conversion, source=origin, target=model_type)
    return d_conv, s_conv


def has_model_origin(cls: Type) -> bool:
    return hasattr(cls, MODEL_ORIGIN_ATTR)


def get_model_origin(cls: Type) -> Type:
    return getattr(cls, MODEL_ORIGIN_ATTR)



from inspect import Parameter, isclass, signature
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    NoReturn,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from apischema.types import AnyType
from apischema.typing import get_type_hints

try:
    from apischema.typing import Annotated, Literal
except ImportError:
    Annotated, Literal = ..., ...  # type: ignore

Converter = Callable[[Any], Any]


def converter_types(
    converter: Converter,
    source: Optional[AnyType] = None,
    target: Optional[AnyType] = None,
    namespace: Dict[str, Any] = None,
) -> Tuple[AnyType, AnyType]:
    try:
        # in pre 3.9, Generic __new__ perturb signature of types
        if (
            isinstance(converter, type)
            and converter.__new__ is Generic.__new__ is not object.__new__
            and converter.__init__ is not object.__init__  # type: ignore
        ):
            parameters = list(signature(converter.__init__).parameters.values())[1:]  # type: ignore # noqa: E501
        else:
            parameters = list(signature(converter).parameters.values())
    except ValueError:  # builtin types
        if target is None and isclass(converter):
            target = cast(Type[Any], converter)
        if source is None:
            raise TypeError("Converter source is unknown") from None
    else:
        if not parameters:
            raise TypeError("converter must have at least one parameter")
        first_param, *other_params = parameters
        for p in other_params:
            if p.default is Parameter.empty and p.kind not in (
                Parameter.VAR_POSITIONAL,
                Parameter.VAR_KEYWORD,
            ):
                raise TypeError(
                    "converter must have at most one parameter without default"
                )
        if source is not None and target is not None:
            return source, target
        types = get_type_hints(converter, None, namespace, include_extras=True)
        if not types and isclass(converter):
            types = get_type_hints(
                converter.__new__, None, namespace, include_extras=True
            ) or get_type_hints(
                converter.__init__, None, namespace, include_extras=True  # type: ignore
            )
        if source is None:
            try:
                source = types.pop(first_param.name)
            except KeyError:
                raise TypeError("converter source is unknown") from None
        if target is None:
            try:
                target = types.pop("return")
            except KeyError:
                if isclass(converter):
                    target = cast(Type, converter)
                else:
                    raise TypeError("converter target is unknown") from None
    return source, target


INVALID_CONVERSION_TYPES = (Union, Annotated, Literal, NoReturn)

T = TypeVar("T")


def identity(x: T) -> T:
    return x



from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    List,
    NewType,
    Optional,
    TYPE_CHECKING,
    Tuple,
    Union,
)

from apischema.conversions.utils import Converter, converter_types, identity
from apischema.dataclasses import replace
from apischema.types import AnyType
from apischema.utils import get_origin_or_type, is_method, method_class, method_wrapper

if TYPE_CHECKING:
    from apischema.deserialization.coercion import Coercion


@dataclass(frozen=True)
class Conversion:
    converter: Union[Converter, property]
    source: AnyType = None
    target: AnyType = None
    sub_conversions: Optional["Conversions"] = None
    additional_properties: Optional[bool] = None
    coercion: Optional["Coercion"] = None
    default_fallback: Optional[bool] = None
    exclude_unset: Optional[bool] = None

    def __call__(self, *args, **kwargs):
        return self.converter(*args, **kwargs)


@dataclass(frozen=True)
class LazyConversion:
    get: Callable[[], Optional["Conversions"]]


ConvOrFunc = Union[Conversion, Converter, property, LazyConversion]
Conversions = Union[ConvOrFunc, Collection[ConvOrFunc]]

ResolvedConversion = NewType("ResolvedConversion", Conversion)
ResolvedConversions = Tuple[ResolvedConversion, ...]  # Tuple in order to be hashable


def resolve_conversion(
    conversion: Union[Converter, property, Conversion], namespace: Dict[str, Any] = None
) -> ResolvedConversion:
    if not isinstance(conversion, Conversion):
        conversion = Conversion(conversion)
    if is_method(conversion.converter):
        if conversion.source is None:
            conversion = replace(conversion, source=method_class(conversion.converter))
        conversion = replace(conversion, converter=method_wrapper(conversion.converter))
    assert not isinstance(conversion.converter, property)
    source, target = converter_types(
        conversion.converter, conversion.source, conversion.target, namespace
    )
    return ResolvedConversion(replace(conversion, source=source, target=target))


def resolve_conversions(conversions: Optional[Conversions]) -> ResolvedConversions:
    if not conversions:
        return ()
    result: List[ResolvedConversion] = []
    for conv in conversions if isinstance(conversions, Collection) else [conversions]:
        if isinstance(conv, LazyConversion):
            result.extend(resolve_conversions(conv.get()))  # type: ignore
        else:
            result.append(resolve_conversion(conv))
    return tuple(result)


def is_identity(conversion: ResolvedConversion) -> bool:
    return (
        conversion.converter == identity
        and conversion.source == conversion.target
        and conversion.sub_conversions is None
        and conversion.additional_properties is None
        and conversion.coercion is None
        and conversion.default_fallback is None
        and conversion.exclude_unset is None
    )


def handle_container_conversions(
    tp: AnyType,
    next_conversions: Optional[Conversions],
    prev_conversions: Optional[Conversions],
    dynamic: bool,
) -> Optional[Conversions]:
    origin = get_origin_or_type(tp)
    if (
        prev_conversions
        and not dynamic
        and (
            (isinstance(origin, type) and issubclass(origin, Collection))
            or origin == Union
        )
    ):
        if next_conversions:
            return (
                LazyConversion(lambda: next_conversions),
                LazyConversion(lambda: prev_conversions),
            )
        else:
            return prev_conversions
    else:
        return next_conversions



from functools import wraps
from inspect import Parameter, iscoroutinefunction, signature
from typing import (
    Awaitable,
    Callable,
    Collection,
    Iterator,
    List,
    NewType,
    Optional,
    Tuple,
    Type,
    TypeVar,
)

from dataclasses import Field, MISSING, field, make_dataclass
from graphql.pyutils import camel_to_snake

from apischema.aliases import alias
from apischema.json_schema.refs import schema_ref
from apischema.utils import Undefined
from apischema.graphql.schema import Mutation as Mutation_
from apischema.graphql.resolvers import awaitable_origin
from apischema.json_schema.schema import Schema
from apischema.serialization.serialized_methods import ErrorHandler
from apischema.types import AnyType
from apischema.typing import get_type_hints
from apischema.utils import get_origin_or_type, is_union_of

ClientMutationId = NewType("ClientMutationId", str)
schema_ref(None)(ClientMutationId)
CLIENT_MUTATION_ID = "client_mutation_id"
M = TypeVar("M", bound="Mutation")


class Mutation:
    _error_handler: ErrorHandler = Undefined
    _schema: Optional[Schema] = None
    _client_mutation_id: Optional[bool] = None
    _mutation: Mutation_  # set in __init_subclass__

    # Mutate is not defined to prevent Mypy warning about signature of superclass
    mutate: Callable

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if not hasattr(cls, "mutate"):
            return
        if not isinstance(cls.__dict__["mutate"], (classmethod, staticmethod)):
            raise TypeError(f"{cls.__name__}.mutate must be a classmethod/staticmethod")
        mutate = getattr(cls, "mutate")
        schema_ref(f"{cls.__name__}Payload")(cls)
        types = get_type_hints(mutate, localns={cls.__name__: cls}, include_extras=True)
        async_mutate = (
            iscoroutinefunction(mutate)
            or get_origin_or_type(types.get("return")) == awaitable_origin
        )
        fields: List[Tuple[str, AnyType, Field]] = []
        cmi_param = None
        for param_name, param in signature(mutate).parameters.items():
            if param.kind is Parameter.POSITIONAL_ONLY:
                raise TypeError("Positional only parameters are not supported")
            if param.kind in {Parameter.POSITIONAL_OR_KEYWORD, Parameter.KEYWORD_ONLY}:
                if param_name not in types:
                    raise TypeError("Mutation parameters must be typed")
                field_type = types[param_name]
                field_ = MISSING if param.default is Parameter.empty else param.default
                if is_union_of(field_type, ClientMutationId):
                    cmi_param = param_name
                    if cls._client_mutation_id is False:
                        if field_ is MISSING:
                            raise TypeError(
                                "Cannot have a ClientMutationId parameter"
                                " when _client_mutation_id = False"
                            )
                        continue
                    elif cls._client_mutation_id is True:
                        field_ = MISSING
                    field_ = field(default=field_, metadata=alias(CLIENT_MUTATION_ID))
                fields.append((param_name, field_type, field_))
        field_names = [name for (name, _, _) in fields]
        if cmi_param is None and cls._client_mutation_id is not False:
            fields.append(
                (
                    CLIENT_MUTATION_ID,
                    ClientMutationId
                    if cls._client_mutation_id
                    else Optional[ClientMutationId],
                    MISSING if cls._client_mutation_id else None,
                )
            )
            cmi_param = CLIENT_MUTATION_ID
        input_cls = make_dataclass(f"{cls.__name__}Input", fields)

        def wrapper(input):
            return mutate(**{name: getattr(input, name) for name in field_names})

        wrapper.__annotations__["input"] = input_cls
        wrapper.__annotations__["return"] = Awaitable[cls] if async_mutate else cls
        if cls._client_mutation_id is not False:
            cls.__annotations__[CLIENT_MUTATION_ID] = input_cls.__annotations__[
                cmi_param
            ]
            setattr(cls, CLIENT_MUTATION_ID, field(init=False))
            wrapped = wrapper

            if async_mutate:

                async def wrapper(input):
                    result = await wrapped(input)
                    setattr(result, CLIENT_MUTATION_ID, getattr(input, cmi_param))
                    return result

            else:

                def wrapper(input):
                    result = wrapped(input)
                    setattr(result, CLIENT_MUTATION_ID, getattr(input, cmi_param))
                    return result

            wrapper = wraps(wrapped)(wrapper)

        cls._mutation = Mutation_(
            function=wrapper,
            alias=camel_to_snake(cls.__name__),
            schema=cls._schema,
            error_handler=cls._error_handler,
        )


def _mutations(cls: Type[Mutation] = Mutation) -> Iterator[Type[Mutation]]:
    for base in cls.__subclasses__():
        if hasattr(base, "_mutation"):
            yield base
            yield from _mutations(base)


def mutations() -> Collection[Mutation_]:
    return [mut._mutation for mut in _mutations()]



import re
import sys
from base64 import b64decode, b64encode
from collections import deque
from datetime import date, datetime, time
from decimal import Decimal
from ipaddress import (
    IPv4Address,
    IPv4Interface,
    IPv4Network,
    IPv6Address,
    IPv6Interface,
    IPv6Network,
)
from pathlib import (
    Path,
    PosixPath,
    PurePath,
    PurePosixPath,
    PureWindowsPath,
    WindowsPath,
)
from typing import Deque, List, NewType, TypeVar
from uuid import UUID

from apischema.conversions import Conversion, as_str, deserializer, serializer
from apischema.json_schema.schema import schema

T = TypeVar("T")


# =================== bytes =====================

deserializer(Conversion(b64decode, source=str, target=bytes))


@serializer
def to_base64(b: bytes) -> str:
    return b64encode(b).decode()


# ================ collections ==================

deserializer(Conversion(deque, source=List[T], target=Deque[T]))
serializer(Conversion(list, source=Deque[T], target=List[T]))
if sys.version_info < (3, 7):
    deserializer(Conversion(deque, source=List, target=deque))
    serializer(Conversion(list, source=deque, target=List))


# ================== datetime ===================

if sys.version_info >= (3, 7):  # pragma: no cover
    for cls, format in [(date, "date"), (datetime, "date-time"), (time, "time")]:
        str_type = schema(format=format)(NewType(cls.__name__.capitalize(), str))
        deserializer(Conversion(cls.fromisoformat, source=str_type, target=cls))  # type: ignore # noqa: E501
        serializer(Conversion(cls.isoformat, source=cls, target=str_type))  # type: ignore # noqa: E501

# ================== decimal ====================

deserializer(Conversion(Decimal, source=float, target=Decimal))
serializer(Conversion(float, source=Decimal, target=float))

# ================= ipaddress ===================

for cls in (IPv4Address, IPv4Interface, IPv4Network):
    schema(format="ipv4")(as_str(cls))
for cls in (IPv6Address, IPv6Interface, IPv6Network):
    schema(format="ipv6")(as_str(cls))

# ==================== path =====================

for cls in (PurePath, PurePosixPath, PureWindowsPath, Path, PosixPath, WindowsPath):
    as_str(cls)

# =================== pattern ===================

Pattern = type(re.compile(r""))
deserializer(Conversion(re.compile, source=str, target=Pattern))
serializer(Conversion(lambda p: p.pattern, source=Pattern, target=str))


# ==================== uuid =====================

schema(format="uuid")(as_str(UUID))



from contextlib import contextmanager
from dataclasses import Field, dataclass
from enum import Enum
from functools import wraps
from itertools import chain
from typing import (  # type: ignore
    AbstractSet,
    Any,
    Callable,
    Collection,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from apischema import settings
from apischema.aliases import Aliaser
from apischema.conversions import identity
from apischema.conversions.conversions import Conversions
from apischema.conversions.utils import Converter
from apischema.conversions.visitor import (
    Conv,
    ConversionsVisitor,
    DeserializationVisitor,
    SerializationVisitor,
)
from apischema.dataclass_utils import (
    get_alias,
    get_default,
    get_field_conversions,
    get_fields,
    get_requirements,
    has_default,
    is_dataclass,
    is_required,
)
from apischema.dataclasses import replace
from apischema.dependent_required import DependentRequired
from apischema.json_schema.constraints import (
    ArrayConstraints,
    Constraints,
    NumberConstraints,
    ObjectConstraints,
    StringConstraints,
)
from apischema.json_schema.generation.refs import Refs, RefsExtractor
from apischema.json_schema.patterns import infer_pattern
from apischema.json_schema.refs import get_ref, schema_ref
from apischema.json_schema.schema import Schema, get_schema, merge_schema
from apischema.json_schema.types import JsonSchema, JsonType, json_schema
from apischema.json_schema.versions import JsonSchemaVersion, RefFactory
from apischema.metadata.keys import (
    MERGED_METADATA,
    PROPERTIES_METADATA,
    SCHEMA_METADATA,
    check_metadata,
)
from apischema.serialization import serialize
from apischema.serialization.serialized_methods import get_serialized_methods
from apischema.skip import filter_skipped
from apischema.tagged_unions import TAGS_ATTR
from apischema.types import AnyType, OrderedDict
from apischema.typing import get_origin
from apischema.utils import UndefinedType, is_union_of, sort_by_annotations_position

constraint_by_type = {
    int: NumberConstraints,
    float: NumberConstraints,
    str: StringConstraints,
}


Method = TypeVar("Method", bound=Callable[..., JsonSchema])


def with_schema(method: Method) -> Method:
    @wraps(method)
    def wrapper(self: "SchemaBuilder", *args, **kwargs):
        if self._schema is None:
            return method(self, *args, **kwargs)
        schema = self._schema.as_dict()
        if not self._schema.override:
            schema = {**method(self, *args, **kwargs), **schema}
        return JsonSchema(schema)

    return cast(Method, wrapper)


@dataclass(frozen=True)
class ObjectField:
    name: str
    alias: str
    required: bool
    schema: JsonSchema


class SchemaBuilder(ConversionsVisitor[Conv, JsonSchema]):
    def __init__(
        self,
        aliaser: Aliaser,
        ref_factory: RefFactory,
        refs: Collection[str],
        ignore_first_ref: bool,
    ):
        super().__init__()
        self.ref_factory = ref_factory
        self.refs = refs
        self.aliaser = aliaser
        self._ignore_first_ref = ignore_first_ref
        self._schema: Optional[Schema] = None

    def _check_constraints(self, expected: Type[Constraints]):
        if self._schema is not None and self._schema.constraints is not None:
            if not isinstance(self._schema.constraints, expected):
                raise TypeError(
                    f"Bad constraints: expected {expected.__name__}"
                    f" found {type(self._schema.constraints).__name__}"
                )

    def _merge_schema(self, schema: Optional[Schema]):
        self._schema = merge_schema(schema, self._schema)

    @with_schema
    def _ref_schema(self, ref: str) -> JsonSchema:
        return JsonSchema({"$ref": self.ref_factory(ref)})

    def annotated(self, tp: AnyType, annotations: Sequence[Any]) -> JsonSchema:
        for annotation in reversed(annotations):
            if isinstance(annotation, schema_ref):
                annotation.check_type(tp)
                if annotation.ref in self.refs:
                    if self._ignore_first_ref:
                        self._ignore_first_ref = False
                    else:
                        assert isinstance(annotation.ref, str)
                        return self._ref_schema(annotation.ref)
                ref = annotation.ref
                if not isinstance(ref, str):
                    raise ValueError("Annotated schema_ref can only be str")
            if isinstance(annotation, Mapping) and SCHEMA_METADATA in annotation:
                self._merge_schema(annotation[SCHEMA_METADATA])
        return self.visit_with_schema(tp, self._schema)

    @with_schema
    def any(self) -> JsonSchema:
        return JsonSchema()

    @with_schema
    def collection(self, cls: Type[Iterable], value_type: AnyType) -> JsonSchema:
        self._check_constraints(ArrayConstraints)
        return json_schema(
            type=JsonType.ARRAY,
            items=self.visit(value_type),
            uniqueItems=issubclass(cls, AbstractSet),
        )

    def visit_field(self, field: Field, field_type: AnyType):
        conversions = get_field_conversions(field, self.operation)
        converter: Converter = identity
        schema: Optional[Schema] = None
        if SCHEMA_METADATA in field.metadata:
            schema = cast(Schema, field.metadata[SCHEMA_METADATA])
            if schema.annotations is not None and schema.annotations is ...:
                if not has_default(field):
                    raise TypeError("Invalid ... without field default")
                try:
                    default = serialize(
                        converter(get_default(field)), conversions=conversions
                    )
                except Exception:
                    pass
                else:
                    annotations = replace(schema.annotations, default=default)
                    schema = replace(schema, annotations=annotations)
        with self._replace_conversions(conversions):
            return self.visit_with_schema(field_type, schema)

    def _properties_schema(self, field: Field, field_type: AnyType) -> JsonSchema:
        with self._without_ref():
            props_schema = self.visit_field(field, field_type)
        if not props_schema.get("type") == JsonType.OBJECT:
            raise TypeError("properties field must have an 'object' type")
        if "patternProperties" in props_schema:
            if (
                len(props_schema["patternProperties"]) != 1
                or "additionalProperties" in props_schema
            ):  # don't try to merge the schemas
                pass
            else:
                return next(iter(props_schema["patternProperties"].values()))
        elif "additionalProperties" in props_schema:
            if isinstance(props_schema["additionalProperties"], JsonSchema):
                return props_schema["additionalProperties"]
            else:  # there is maybe only properties
                pass
        return JsonSchema()

    def _check_merged_schema(self, cls: Type, field: Field, field_type: AnyType):
        ignore_first_ref = self._ignore_first_ref
        self._ignore_first_ref = True
        try:
            if self.visit_field(field, field_type).get("type") not in {
                JsonType.OBJECT,
                "object",
            }:
                raise TypeError(
                    f"Merged field {cls.__name__}.{field.name}"
                    f" must have an object type"
                )
        finally:
            self._ignore_first_ref = ignore_first_ref

    def _update_properties(
        self, cls: Type, properties: Dict[str, JsonSchema], required: List[str]
    ):
        pass

    def _properties(
        self, cls: Type, fields: Collection[ObjectField]
    ) -> Tuple[Mapping[str, JsonSchema], Sequence[str]]:
        properties, required = {}, []
        for field in fields:
            alias = field.alias or self.aliaser(field.name)
            properties[alias] = field.schema
            if field.required and alias not in required:
                required.append(alias)
        return properties, required

    @with_schema
    def dataclass(
        self,
        cls: Type,
        types: Mapping[str, AnyType],
        fields: Sequence[Field],
        init_vars: Sequence[Field],
    ) -> JsonSchema:
        assert is_dataclass(cls)
        self._check_constraints(ObjectConstraints)
        merged_schemas: List[JsonSchema] = []
        pattern_properties = {}
        additional_properties: Union[bool, JsonSchema] = settings.additional_properties
        object_fields: List[ObjectField] = []
        for field in get_fields(fields, init_vars, self.operation):
            metadata = check_metadata(field)
            field_type = types[field.name]
            if MERGED_METADATA in metadata:
                self._check_merged_schema(cls, field, field_type)
                merged_schemas.append(self.visit_field(field, field_type))
            elif PROPERTIES_METADATA in metadata:
                pattern = metadata[PROPERTIES_METADATA]
                properties_schema = self._properties_schema(field, field_type)
                if pattern is None:
                    additional_properties = properties_schema
                elif pattern is ...:
                    pattern_properties[infer_pattern(field_type)] = properties_schema
                else:
                    pattern_properties[pattern] = properties_schema
            else:
                alias = self.aliaser(get_alias(field))
                field_schema = json_schema(
                    readOnly=not field.init,
                    writeOnly=field in init_vars,
                    **self.visit_field(field, field_type),
                )
                object_field = ObjectField(
                    field.name, alias, is_required(field), field_schema
                )
                object_fields.append(object_field)
        properties, required = self._properties(cls, object_fields)
        dependent_required = {
            self.aliaser(get_alias(field)): sorted(
                self.aliaser(get_alias(req)) for req in required_by
            )
            for field, required_by in get_requirements(
                cls, DependentRequired.requiring, self.operation
            ).items()
        }
        if hasattr(cls, TAGS_ATTR):
            return json_schema(
                type=JsonType.OBJECT,
                oneOf=[
                    json_schema(properties={prop: value})
                    for prop, value in properties.items()
                ],
            )
        result = json_schema(
            type=JsonType.OBJECT,
            properties=properties,
            required=required,
            additionalProperties=additional_properties,
            patternProperties=pattern_properties,
            dependentRequired=OrderedDict(
                (f, dependent_required[f]) for f in sorted(dependent_required)
            ),
        )
        if merged_schemas:
            result = json_schema(
                type=JsonType.OBJECT,
                allOf=[result, *merged_schemas],
                unevaluatedProperties=False,
            )
        return result

    def enum(self, cls: Type[Enum]) -> JsonSchema:
        if len(cls) == 0:
            raise TypeError("Empty enum")
        return self.literal(list(cls))

    def generic(self, tp: AnyType) -> JsonSchema:
        self._merge_schema(get_schema(get_origin(tp)))
        return super().generic(tp)

    @with_schema
    def literal(self, values: Sequence[Any]) -> JsonSchema:
        if not values:
            raise TypeError("Empty Literal")
        types = {
            JsonType.from_type(type(v.value if isinstance(v, Enum) else v))
            for v in values
        }
        # Mypy issue
        type_: Any = types.pop() if len(types) == 1 else types
        if len(values) == 1:
            return json_schema(type=type_, const=values[0])
        else:
            return json_schema(type=type_, enum=values)

    @with_schema
    def mapping(
        self,
        cls: Type[Mapping],
        key_type: AnyType,
        value_type: AnyType,
    ) -> JsonSchema:
        self._check_constraints(ObjectConstraints)
        with self._without_ref():
            key = self.visit(key_type)
        if key["type"] != JsonType.STRING:
            raise ValueError("Mapping types must string-convertible key")
        value = self.visit(value_type)
        if "pattern" in key:
            return json_schema(
                type=JsonType.OBJECT, patternProperties={key["pattern"]: value}
            )
        else:
            return json_schema(type=JsonType.OBJECT, additionalProperties=value)

    @with_schema
    def named_tuple(
        self,
        cls: Type[Tuple],
        types: Mapping[str, AnyType],
        defaults: Mapping[str, Any],
    ) -> JsonSchema:
        self._check_constraints(ObjectConstraints)
        properties = {self.aliaser(key): self.visit(key) for key, cls in types.items()}
        required = [attr for attr in types if attr not in defaults]
        self._update_properties(cls, properties, required)
        return json_schema(
            type=JsonType.OBJECT,
            properties=properties,
            required=required,
            additionalProperties=settings.additional_properties,
        )

    def new_type(self, tp: Type, super_type: AnyType) -> JsonSchema:
        return self.visit_with_schema(super_type, self._schema)

    @with_schema
    def primitive(self, cls: Type) -> JsonSchema:
        if cls in constraint_by_type:
            self._check_constraints(constraint_by_type[cls])
        return JsonSchema(type=JsonType.from_type(cls))

    def subprimitive(self, cls: Type, superclass: Type) -> JsonSchema:
        return self.primitive(superclass)

    @with_schema
    def tuple(self, types: Sequence[AnyType]) -> JsonSchema:
        self._check_constraints(ArrayConstraints)
        if self._schema is not None and self._schema.constraints is not None:
            assert isinstance(self._schema.constraints, ArrayConstraints)
            if (
                self._schema.constraints.max_items is not None
                or self._schema.constraints.min_items is not None
            ):
                raise TypeError("Tuple cannot have min_items/max_items constraints")
        return json_schema(
            type=JsonType.ARRAY,
            items=[self.visit(cls) for cls in types],
            minItems=len(types),
            maxItems=len(types),
        )

    @with_schema
    def typed_dict(
        self, cls: Type, keys: Mapping[str, AnyType], total: bool
    ) -> JsonSchema:
        self._check_constraints(ObjectConstraints)
        return json_schema(
            type=JsonType.OBJECT,
            properties={key: self.visit(tp) for key, tp in keys.items()},
            required=list(keys) if total else [],
        )

    def _union_result(self, results: Iterable[JsonSchema]) -> JsonSchema:
        results = list(results)
        if len(results) == 1:
            return results[0]
        elif all(alt.keys() == {"type"} for alt in results):
            types: Any = chain.from_iterable(
                [res["type"]]
                if isinstance(res["type"], (str, JsonType))
                else res["type"]
                for res in results
            )
            return json_schema(type=list(types))
        elif (
            len(results) == 2
            and all("type" in res for res in results)
            and {"type": "null"} in results
        ):
            for result in results:
                if result != {"type": "null"}:
                    types = result["type"]
                    if isinstance(types, (str, JsonType)):
                        types = [types]
                    if "null" not in types:
                        result = JsonSchema({**result, "type": [*types, "null"]})
                    return result
            else:
                raise NotImplementedError
        else:
            return json_schema(anyOf=results)

    @with_schema
    def union(self, alternatives: Sequence[AnyType]) -> JsonSchema:
        return self._union_result(
            map(self.visit, filter_skipped(alternatives, schema_only=True))
        )

    def visit_with_schema(self, tp: AnyType, schema: Optional[Schema]) -> JsonSchema:
        schema_save = self._schema
        dynamic = self._apply_dynamic_conversions(tp)
        tp, self._schema = (dynamic, None) if dynamic is not None else (tp, schema)
        ref = get_ref(tp)
        if ref in self.refs:
            if self._ignore_first_ref:
                self._ignore_first_ref = False
            else:
                assert isinstance(ref, str)
                return self._ref_schema(ref)
        self._merge_schema(get_schema(tp))
        try:
            return super().visit(tp)
        finally:
            self._schema = schema_save

    def visit(self, tp: AnyType) -> JsonSchema:
        return self.visit_with_schema(tp, None)

    @contextmanager
    def _without_ref(self):
        refs_save = self.refs
        self.refs = ()
        try:
            yield
        finally:
            self.refs = refs_save

    RefsExtractor: Type["RefsExtractor"]


class DeserializationSchemaBuilder(DeserializationVisitor, SchemaBuilder):
    class RefsExtractor(DeserializationVisitor, RefsExtractor):  # type: ignore
        pass

    visit_conversion = with_schema(DeserializationVisitor.visit_conversion)  # type: ignore # noqa: E501


class SerializationSchemaBuilder(SerializationVisitor, SchemaBuilder):
    class RefsExtractor(SerializationVisitor, RefsExtractor):  # type: ignore
        pass

    def _properties(
        self, cls: Type, fields: Collection[ObjectField]
    ) -> Tuple[Mapping[str, JsonSchema], Sequence[str]]:
        all_fields = list(fields)
        for name, (serialized, types) in get_serialized_methods(
            self._generic or cls
        ).items():
            ret = types["return"]
            with self._replace_conversions(serialized.conversions):
                schema = json_schema(
                    readOnly=True, **self.visit_with_schema(ret, serialized.schema)
                )
            required = not is_union_of(ret, UndefinedType)
            all_fields.append(
                ObjectField(serialized.func.__name__, name, required, schema)
            )
        all_fields = sort_by_annotations_position(cls, all_fields, lambda f: f.name)
        return super()._properties(cls, all_fields)

    visit_conversion = with_schema(SerializationVisitor.visit_conversion)  # type: ignore # noqa: E501


TypesWithConversions = Collection[Union[AnyType, Tuple[AnyType, Conversions]]]


def _default_version(
    version: Optional[JsonSchemaVersion],
    ref_factory: Optional[RefFactory],
    all_refs: Optional[bool],
) -> Tuple[JsonSchemaVersion, RefFactory, bool]:
    if version is None:
        version = settings.json_schema_version
    if ref_factory is None:
        ref_factory = version.ref_factory
    if all_refs is None:
        all_refs = version.all_refs
    return version, ref_factory, all_refs


def _extract_refs(
    types: TypesWithConversions, builder: Type[SchemaBuilder], all_refs: bool
) -> Mapping[str, AnyType]:
    refs: Refs = {}
    for cls in types:
        conversions = None
        if isinstance(cls, tuple):
            cls, conversions = cls
        builder.RefsExtractor(refs).visit_with_conversions(cls, conversions)
    filtr = (lambda count: True) if all_refs else (lambda count: count > 1)
    return {ref: cls for ref, (cls, count) in refs.items() if filtr(count)}


def _refs_schema(
    builder: Type[SchemaBuilder],
    aliaser: Aliaser,
    refs: Mapping[str, AnyType],
    ref_factory: RefFactory,
) -> Mapping[str, JsonSchema]:
    return {
        ref: builder(aliaser, ref_factory, refs, True).visit(cls)
        for ref, cls in refs.items()
    }


def _schema(
    builder: Type[SchemaBuilder],
    tp: AnyType,
    schema: Optional[Schema],
    conversions: Optional[Conversions],
    version: Optional[JsonSchemaVersion],
    aliaser: Optional[Aliaser],
    ref_factory: Optional[RefFactory],
    all_refs: Optional[bool],
    with_schema: bool,
) -> Mapping[str, Any]:
    add_defs = ref_factory is None
    if ref_factory is not None and all_refs is None:
        all_refs = True
    if aliaser is None:
        aliaser = settings.aliaser()
    version, ref_factory, all_refs = _default_version(version, ref_factory, all_refs)
    refs = _extract_refs([(tp, conversions)], builder, all_refs)
    visitor = builder(aliaser, ref_factory, refs, False)
    with visitor._replace_conversions(conversions):
        json_schema = visitor.visit_with_schema(tp, schema)
    if add_defs:
        defs = _refs_schema(builder, aliaser, refs, ref_factory)
        if defs:
            json_schema["$defs"] = defs
    result = serialize(json_schema, conversions=version.conversion)
    if with_schema and version.schema is not None:
        result["$schema"] = version.schema
    return result


def deserialization_schema(
    tp: AnyType,
    *,
    schema: Schema = None,
    conversions: Conversions = None,
    version: JsonSchemaVersion = None,
    aliaser: Aliaser = None,
    ref_factory: RefFactory = None,
    all_refs: bool = None,
    with_schema: bool = True,
) -> Mapping[str, Any]:
    return _schema(
        DeserializationSchemaBuilder,
        tp,
        schema,
        conversions,
        version,
        aliaser,
        ref_factory,
        all_refs,
        with_schema,
    )


def serialization_schema(
    tp: AnyType,
    *,
    schema: Schema = None,
    conversions: Conversions = None,
    version: JsonSchemaVersion = None,
    aliaser: Aliaser = None,
    ref_factory: RefFactory = None,
    all_refs: bool = None,
    with_schema: bool = True,
) -> Mapping[str, Any]:
    return _schema(
        SerializationSchemaBuilder,
        tp,
        schema,
        conversions,
        version,
        aliaser,
        ref_factory,
        all_refs,
        with_schema,
    )


def _defs_schema(
    types: TypesWithConversions,
    builder: Type[SchemaBuilder],
    aliaser: Aliaser,
    ref_factory: RefFactory,
    all_refs: bool,
) -> Mapping[str, JsonSchema]:
    return _refs_schema(
        builder, aliaser, _extract_refs(types, builder, all_refs), ref_factory
    )


def _set_missing_properties(
    schema: JsonSchema, properties: Optional[Mapping[str, JsonSchema]], key: str
) -> JsonSchema:
    if properties is None:
        return schema
    missing = {name: prop for name, prop in properties.items() if prop.get(key, False)}
    schema.setdefault("properties", {}).update(missing)
    return schema


def definitions_schema(
    *,
    deserialization: TypesWithConversions = (),
    serialization: TypesWithConversions = (),
    aliaser: Aliaser = None,
    version: JsonSchemaVersion = None,
    ref_factory: Optional[RefFactory] = None,
    all_refs: bool = None,
) -> Mapping[str, Mapping[str, Any]]:
    if aliaser is None:
        aliaser = settings.aliaser()
    version, ref_factory, all_refs = _default_version(version, ref_factory, all_refs)
    deserialization_schemas = _defs_schema(
        deserialization, DeserializationSchemaBuilder, aliaser, ref_factory, all_refs
    )
    serialization_schemas = _defs_schema(
        serialization, SerializationSchemaBuilder, aliaser, ref_factory, all_refs
    )
    for duplicate in deserialization_schemas.keys() & serialization_schemas.keys():
        d_schema = deserialization_schemas[duplicate]
        s_schema = serialization_schemas[duplicate]
        _set_missing_properties(s_schema, d_schema.get("properties"), "writeOnly")
        _set_missing_properties(d_schema, s_schema.get("properties"), "readOnly")
        if "required" in d_schema and "required" in s_schema:
            s_schema["required"] = d_schema["required"]
        if d_schema != s_schema:
            raise TypeError(
                f"Reference {duplicate} has different schemas"
                f" for deserialization and serialization"
            )
    return {
        ref: serialize(schema, conversions=version.conversion)
        for ref, schema in chain(
            deserialization_schemas.items(), serialization_schemas.items()
        )
    }



from collections import defaultdict
from dataclasses import dataclass, is_dataclass
from enum import Enum
from functools import partial
from inspect import Parameter, iscoroutinefunction, signature
from typing import (
    Any,
    Awaitable,
    Callable,
    Collection,
    Dict,
    Iterator,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    overload,
)

import graphql

from apischema.aliases import Aliaser
from apischema.conversions.conversions import Conversions, handle_container_conversions
from apischema.conversions.dataclass_models import DataclassModel
from apischema.deserialization import deserialize
from apischema.json_schema.schema import Schema
from apischema.metadata.implem import ConversionMetadata
from apischema.metadata.keys import (
    ALIAS_METADATA,
    CONVERSIONS_METADATA,
    DEFAULT_FALLBACK_METADATA,
    REQUIRED_METADATA,
    get_annotated_metadata,
)
from apischema.serialization import (
    COLLECTION_TYPE_SET,
    MAPPING_TYPE_SET,
    PRIMITIVE_TYPES_SET,
    get_conversions,
    serialize,
)
from apischema.serialization.serialized_methods import (
    ErrorHandler,
    Serialized,
    _get_methods,
    serialized as register_serialized,
)
from apischema.types import AnyType, NoneType, PRIMITIVE_TYPES
from apischema.typing import get_origin, get_type_hints
from apischema.utils import (
    Undefined,
    get_args2,
    get_origin_or_type,
    is_union_of,
    method_registerer,
)
from apischema.validation.errors import ValidationError
from apischema.visitor import Unsupported


def partial_serialize(
    obj: Any, *, aliaser: Aliaser, conversions: Conversions = None
) -> Any:
    if obj is Undefined:
        return None
    cls = obj.__class__
    if cls in PRIMITIVE_TYPES_SET:
        return obj
    if cls in COLLECTION_TYPE_SET:
        return [
            partial_serialize(elt, aliaser=aliaser, conversions=conversions)
            for elt in obj
        ]
    if cls in MAPPING_TYPE_SET:
        return serialize(
            obj, conversions=conversions, aliaser=aliaser, exclude_unset=False
        )
    conversion, dynamic = get_conversions(cls, conversions)
    if conversion is not None:
        if isinstance(conversion.target, DataclassModel):
            return obj
        return partial_serialize(
            conversion.converter(obj),  # type: ignore
            aliaser=aliaser,
            conversions=handle_container_conversions(
                conversion.target,
                conversion.sub_conversions,
                conversions,
                dynamic,
            ),
        )
    if is_dataclass(cls):
        return obj
    if issubclass(cls, Enum):
        return serialize(obj.value, aliaser=aliaser, exclude_unset=False)
    if isinstance(obj, PRIMITIVE_TYPES):
        return obj
    if isinstance(obj, Mapping):
        return serialize(obj, aliaser=aliaser, exclude_unset=False)
    if isinstance(obj, Collection):
        return [partial_serialize(elt, aliaser=aliaser) for elt in obj]
    if issubclass(cls, tuple) and hasattr(cls, "_fields"):
        return obj
    raise Unsupported(cls)


awaitable_origin = get_origin(Awaitable[Any])


def is_async(func: Callable) -> bool:
    return (
        iscoroutinefunction(func)
        or get_origin_or_type(get_type_hints(func).get("return")) == awaitable_origin
    )


def unwrap_awaitable(tp: AnyType) -> AnyType:
    return get_args2(tp)[0] if get_origin_or_type(tp) == awaitable_origin else tp


@dataclass(frozen=True)
class Resolver(Serialized):
    parameters: Sequence[Parameter]
    parameters_metadata: Mapping[str, Mapping]

    def error_type(self) -> AnyType:
        return unwrap_awaitable(super().error_type())

    def return_type(self, return_type: AnyType) -> AnyType:
        return super().return_type(unwrap_awaitable(return_type))


_resolvers: Dict[Type, Dict[str, Resolver]] = defaultdict(dict)


def get_resolvers(tp: AnyType) -> Mapping[str, Tuple[Resolver, Mapping[str, AnyType]]]:
    return _get_methods(tp, _resolvers)


def none_error_handler(
    __error: Exception, __obj: Any, __info: graphql.GraphQLResolveInfo, **kwargs
) -> None:
    return None


def resolver_parameters(
    resolver: Callable, *, check_first: bool
) -> Iterator[Parameter]:
    first = True
    for param in signature(resolver).parameters.values():
        if param.kind is Parameter.POSITIONAL_ONLY:
            raise TypeError("Resolver can not have positional only parameters")
        if param.kind in {Parameter.POSITIONAL_OR_KEYWORD, Parameter.KEYWORD_ONLY}:
            if param.annotation is Parameter.empty and (check_first or not first):
                raise TypeError("Resolver parameters must be typed")
            yield param
        first = False


MethodOrProp = TypeVar("MethodOrProp", Callable, property)


@overload
def resolver(__method_or_property: MethodOrProp) -> MethodOrProp:
    ...


@overload
def resolver(
    alias: str = None,
    *,
    conversions: Conversions = None,
    schema: Schema = None,
    error_handler: ErrorHandler = Undefined,
    parameters_metadata: Mapping[str, Mapping] = None,
    serialized: bool = False,
    owner: Type = None,
) -> Callable[[MethodOrProp], MethodOrProp]:
    ...


def resolver(
    __arg=None,
    *,
    alias: str = None,
    conversions: Conversions = None,
    schema: Schema = None,
    error_handler: ErrorHandler = Undefined,
    parameters_metadata: Mapping[str, Mapping] = None,
    serialized: bool = False,
    owner: Type = None,
):
    def register(func: Callable, owner: Optional[Type], alias2: str):
        alias2 = alias or alias2
        first_param, *parameters = resolver_parameters(func, check_first=owner is None)
        error_handler2 = error_handler
        if error_handler2 is None:
            error_handler2 = none_error_handler
        elif error_handler2 is Undefined:
            error_handler2 = None
        resolver = Resolver(
            func,
            conversions,
            schema,
            error_handler2,
            parameters,
            parameters_metadata or {},
        )
        if owner is None:
            try:
                owner = get_origin_or_type(get_type_hints(func)[first_param.name])
            except KeyError:
                raise TypeError("First parameter of resolver must be typed") from None
        _resolvers[owner][alias2] = resolver
        if serialized:
            if is_async(func):
                raise TypeError("Async resolver cannot be used as a serialized method")
            try:
                register_serialized(
                    alias=alias2,
                    conversions=conversions,
                    schema=schema,
                    error_handler=error_handler,
                    owner=owner,
                )(func)
            except Exception:
                raise TypeError("Resolver cannot be used as a serialized method")

    if isinstance(__arg, str):
        alias = __arg
        __arg = None
    return method_registerer(__arg, owner, register)


def resolver_resolve(
    resolver: Resolver,
    types: Mapping[str, AnyType],
    aliaser: Aliaser,
    serialized: bool = True,
) -> Callable:
    parameters, info_parameter = [], None
    for param in resolver.parameters:
        param_type = types[param.name]
        if is_union_of(param_type, graphql.GraphQLResolveInfo):
            info_parameter = param.name
        else:
            metadata = get_annotated_metadata(param_type)
            if param.name in resolver.parameters_metadata:
                metadata = {**metadata, **resolver.parameters_metadata[param.name]}
            alias = metadata.get(ALIAS_METADATA, param.name)
            deserializer = partial(
                deserialize,
                param_type,
                conversions=metadata.get(
                    CONVERSIONS_METADATA, ConversionMetadata()
                ).deserialization,
                aliaser=aliaser,
                default_fallback=DEFAULT_FALLBACK_METADATA in metadata or None,
            )
            required = REQUIRED_METADATA in metadata or param.default is Parameter.empty
            opt_param = is_union_of(param_type, NoneType)
            parameters.append(
                (aliaser(alias), param.name, deserializer, opt_param, required)
            )
    func, error_handler = resolver.func, resolver.error_handler

    def no_serialize(result):
        return result

    async def async_serialize(result: Awaitable):
        return partial_serialize(
            await result, aliaser=aliaser, conversions=resolver.conversions
        )

    def sync_serialize(result):
        return partial_serialize(
            result, aliaser=aliaser, conversions=resolver.conversions
        )

    serialize_result: Callable[[Any], Any]
    if not serialized:
        serialize_result = no_serialize
    elif is_async(resolver.func):
        serialize_result = async_serialize
    else:
        serialize_result = sync_serialize
    serialize_error: Optional[Callable[[Any], Any]]
    if error_handler is None:
        serialize_error = None
    elif is_async(error_handler):
        serialize_error = async_serialize
    else:
        serialize_error = sync_serialize

    def resolve(__self, __info, **kwargs):
        values = {}
        errors: Dict[str, ValidationError] = {}
        for alias, param_name, deserializer, opt_param, required in parameters:
            if alias in kwargs:
                if not opt_param and kwargs[param_name] is None:
                    assert not required
                    continue
                try:
                    values[param_name] = deserializer(kwargs[alias])
                except ValidationError as err:
                    errors[aliaser(param_name)] = err
            elif opt_param and required:
                values[param_name] = None

        if errors:
            raise TypeError(serialize(ValidationError(children=errors)))
        if info_parameter:
            values[info_parameter] = __info
        try:
            return serialize_result(func(__self, **values))
        except Exception as error:
            if error_handler is None:
                raise
            assert serialize_error is not None
            return serialize_error(error_handler(error, __self, __info, **kwargs))

    return resolve



from dataclasses import Field, dataclass, is_dataclass
from enum import Enum
from functools import wraps
from itertools import chain
from typing import (
    AbstractSet,
    Any,
    Callable,
    Collection,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Pattern,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

from apischema import settings
from apischema.aliases import Aliaser
from apischema.cache import cache
from apischema.conversions.conversions import (
    Conversions,
    handle_container_conversions,
)
from apischema.conversions.utils import (
    Converter,
    identity,
)
from apischema.conversions.visitor import Deserialization, DeserializationVisitor
from apischema.dataclass_utils import (
    get_alias,
    get_default,
    get_field_conversions,
    get_fields,
    get_requirements,
    has_default,
    is_required,
)
from apischema.dataclasses import replace
from apischema.dependent_required import DependentRequired
from apischema.deserialization.coercion import Coercion, get_coercer
from apischema.deserialization.merged import get_init_merged_alias
from apischema.json_schema.constraints import (
    ArrayConstraints,
    Constraints,
    merge_constraints,
)
from apischema.json_schema.patterns import infer_pattern
from apischema.json_schema.schema import Schema, get_schema
from apischema.metadata.implem import ValidatorsMetadata
from apischema.metadata.keys import (
    DEFAULT_FALLBACK_METADATA,
    MERGED_METADATA,
    POST_INIT_METADATA,
    PROPERTIES_METADATA,
    SCHEMA_METADATA,
    SKIP_METADATA,
    VALIDATORS_METADATA,
    check_metadata,
    is_aggregate_field,
)
from apischema.skip import filter_skipped
from apischema.types import (
    AnyType,
    COLLECTION_TYPES,
    MAPPING_TYPES,
    NoneType,
    OrderedDict,
)
from apischema.typing import get_origin
from apischema.utils import get_origin_or_type, opt_or
from apischema.validation.errors import (
    ErrorKey,
    ValidationError,
    apply_aliaser,
    merge_errors,
)
from apischema.validation.mock import ValidatorMock
from apischema.validation.validator import (
    Validator,
    get_validators,
    validate,
)
from apischema.visitor import Unsupported

DICT_TYPE = get_origin(Dict[Any, Any])
LIST_TYPE = get_origin(List[Any])

MISSING_PROPERTY = ValidationError(["missing property"])
UNEXPECTED_PROPERTY = ValidationError(["unexpected property"])

T = TypeVar("T")


# TODO maybe ctx parameters to deserializers
@dataclass
class DeserializationContext:
    additional_properties: bool
    coercion: Coercion
    default_fallback: bool

    def __post_init__(self):
        self.coercer = get_coercer(self.coercion)

    def merge(
        self,
        additional_properties: bool = None,
        coercion: Coercion = None,
        default_fallback: bool = None,
    ) -> "DeserializationContext":
        if any(
            arg is not None
            for arg in (additional_properties, coercion, default_fallback)
        ):
            return replace(
                self,
                additional_properties=opt_or(
                    additional_properties, self.additional_properties
                ),
                coercion=opt_or(coercion, self.coercion),
                default_fallback=opt_or(default_fallback, self.default_fallback),
            )
        else:
            return self


def get_constraints(tp: AnyType) -> Optional[Constraints]:
    schema = get_schema(tp)
    return schema.constraints if schema is not None else None


DeserializationMethod = Callable[[DeserializationContext, Any], Any]
Factory = Callable[[Optional[Constraints], Sequence[Validator]], DeserializationMethod]


@dataclass(frozen=True)
class DeserializationMethodFactory:
    factory: Factory
    constraints: Optional[Constraints] = None
    validators: Sequence[Validator] = ()

    @property
    def method(self) -> DeserializationMethod:
        return self.factory(self.constraints, self.validators)  # type: ignore

    @staticmethod
    def from_type(tp: AnyType) -> Callable[[Factory], "DeserializationMethodFactory"]:
        return lambda factory: DeserializationMethodFactory(
            factory, get_constraints(tp), get_validators(tp)
        )

    def merge(
        self,
        constraints: Optional[Constraints] = None,
        validators: Sequence[Validator] = (),
    ) -> "DeserializationMethodFactory":
        if constraints is None and not validators:
            return self
        return replace(
            self,
            constraints=merge_constraints(self.constraints, constraints),
            validators=(*self.validators, *validators),
        )


class RecDeserializerMethodFactory:
    def __init__(self):
        self._ref: Optional[DeserializationMethodFactory] = None
        self._constraints: Optional[Constraints] = None
        self._validators: Sequence[Validator] = ()
        self._children: List[RecDeserializerMethodFactory] = []
        self._method: Optional[DeserializationMethod] = None

    def __hash__(self):
        return object.__hash__(self)

    def set_ref(
        self, factory: DeserializationMethodFactory
    ) -> DeserializationMethodFactory:
        self._ref = factory.merge(self._constraints, self._validators)
        for child in self._children:
            child.set_ref(factory)
        return factory

    def merge(
        self,
        constraints: Optional[Constraints] = None,
        validators: Sequence[Validator] = (),
    ) -> "RecDeserializerMethodFactory":
        child = RecDeserializerMethodFactory()
        if self._ref is not None:
            child._ref = self._ref.merge(constraints, validators)
        else:
            child._constraints = merge_constraints(self._constraints, constraints)
            child._validators = (*self._validators, *validators)
        self._children.append(child)
        return child

    @property  # type: ignore
    def method(self) -> DeserializationMethod:
        if self._method is None:

            def method(ctx: DeserializationContext, data: Any) -> Any:
                if self._method is None:
                    assert self._ref is not None
                    self._method = self._ref.method
                return self._method(ctx, data)

            return method
        else:
            return self._method


FieldDeserializer = Callable[
    [
        DeserializationContext,
        Any,
        Dict[str, Any],
        List[str],
        Dict[ErrorKey, ValidationError],
    ],
    None,
]


def field_deserializer(
    field: Field, method: DeserializationMethod, aliaser: Aliaser
) -> FieldDeserializer:
    name = field.name
    alias = aliaser(get_alias(field))
    aggregate = is_aggregate_field(field)
    default = has_default(field)
    default_fallback = DEFAULT_FALLBACK_METADATA in field.metadata

    def deserializer(
        ctx: DeserializationContext,
        data: Any,
        values: Dict[str, Any],
        errors: List[str],
        field_errors: Dict[ErrorKey, ValidationError],
    ):
        try:
            values[name] = method(ctx, data)  # type: ignore
        except ValidationError as err:
            if default and (default_fallback or ctx.default_fallback):
                pass
            elif aggregate:
                errors.extend(err.messages)
                field_errors.update(err.children)
            else:
                field_errors[alias] = err

    return deserializer


def with_validators(
    validators: Sequence[Validator],
) -> Callable[[DeserializationMethod], DeserializationMethod]:
    if not validators:
        return lambda method: method

    def decorator(method: DeserializationMethod) -> DeserializationMethod:
        @wraps(method)
        def wrapper(ctx: DeserializationContext, data: Any) -> Any:
            return validate(method(ctx, data), validators)

        return wrapper

    return decorator


class DeserializationMethodVisitor(
    DeserializationVisitor[DeserializationMethodFactory]
):
    def __init__(self, aliaser: Aliaser):
        super().__init__()
        self._rec_sentinel: Dict[Any, RecDeserializerMethodFactory] = {}
        self.aliaser = aliaser

    def _visit(self, tp: AnyType) -> DeserializationMethodFactory:
        key = self._generic or tp, self._conversions
        if key in self._rec_sentinel:
            return cast(DeserializationMethodFactory, self._rec_sentinel[key])
        else:
            self._rec_sentinel[key] = RecDeserializerMethodFactory()
            factory = super()._visit(tp)
            return self._rec_sentinel.pop(key).set_ref(factory)

    def method(self, cls) -> DeserializationMethod:
        return self.visit(cls).method

    def annotated(
        self, tp: AnyType, annotations: Sequence[Any]
    ) -> DeserializationMethodFactory:
        factory = self.visit(tp)
        for annotation in reversed(annotations):
            if isinstance(annotation, Mapping):
                if SCHEMA_METADATA in annotation:
                    schema: Schema = annotation[SCHEMA_METADATA]
                    factory = factory.merge(constraints=schema.constraints)
                if VALIDATORS_METADATA in annotation:
                    validators: ValidatorsMetadata = annotation[VALIDATORS_METADATA]
                    factory = factory.merge(validators=validators.validators)
        return factory

    def any(self) -> DeserializationMethodFactory:
        @DeserializationMethodFactory
        def factory(
            constraints: Optional[Constraints], validators: Sequence[Validator]
        ) -> DeserializationMethod:
            @with_validators(validators)
            def method(ctx: DeserializationContext, data: Any) -> Any:
                return data if constraints is None else constraints.validate(data)

            return method

        return factory

    def collection(
        self, cls: Type[Iterable], value_type: AnyType
    ) -> DeserializationMethodFactory:
        deserialize_value = self.method(value_type)

        @DeserializationMethodFactory
        def factory(
            constraints: Optional[Constraints], validators: Sequence[Validator]
        ) -> DeserializationMethod:
            @with_validators(validators)
            def method(ctx: DeserializationContext, data: Any) -> Any:
                data = ctx.coercer(list, data)
                elts = []
                elt_errors: Dict[ErrorKey, ValidationError] = {}
                for i, elt in enumerate(data):
                    try:
                        elts.append(deserialize_value(ctx, elt))
                    except ValidationError as err:
                        elt_errors[i] = err
                errors = () if constraints is None else constraints.errors(data)
                if elt_errors or errors:
                    raise ValidationError(errors, elt_errors)
                return elts if cls is LIST_TYPE else COLLECTION_TYPES[cls](elts)

            return method

        return factory

    def dataclass(
        self,
        cls: Type,
        types: Mapping[str, AnyType],
        fields: Sequence[Field],
        init_vars: Sequence[Field],
    ) -> DeserializationMethodFactory:
        assert is_dataclass(cls)
        normal_fields: List[
            Tuple[str, FieldDeserializer, Union[bool, AbstractSet[str]]]
        ] = []
        merged_fields: List[Tuple[AbstractSet[str], FieldDeserializer]] = []
        pattern_fields: List[Tuple[Pattern, FieldDeserializer]] = []
        additional_field: Optional[FieldDeserializer] = None
        post_init_modified = {
            field.name
            for field in chain(fields, init_vars)
            if POST_INIT_METADATA in field.metadata
        }
        defaults: Dict[str, Callable[[], Any]] = {}
        required_by = get_requirements(
            cls, DependentRequired.required_by, self.operation
        )
        for field in get_fields(fields, init_vars, self.operation):  # noqa: F402
            metadata = check_metadata(field)
            if SKIP_METADATA in metadata or not field.init:
                continue
            if has_default(field):
                defaults[field.name] = lambda: get_default(field)
            field_type = types[field.name]
            field_factory = self.visit_with_conversions(
                field_type, get_field_conversions(field, self.operation)
            )
            if SCHEMA_METADATA in metadata:
                field_factory = field_factory.merge(
                    constraints=metadata[SCHEMA_METADATA].constraints
                )
            if VALIDATORS_METADATA in metadata:
                field_factory = field_factory.merge(
                    validators=metadata[VALIDATORS_METADATA].validators
                )
            deserializer = field_deserializer(field, field_factory.method, self.aliaser)
            if MERGED_METADATA in metadata:
                merged_alias = get_init_merged_alias(cls, field, field_type)
                merged_fields.append(
                    (set(map(self.aliaser, merged_alias)), deserializer)
                )
            elif PROPERTIES_METADATA in metadata:
                pattern = metadata[PROPERTIES_METADATA]
                if pattern is None:
                    additional_field = deserializer
                elif pattern is ...:
                    pattern_fields.append((infer_pattern(field_type), deserializer))
                else:
                    pattern_fields.append((pattern, deserializer))
            else:
                required = is_required(field) or {
                    self.aliaser(get_alias(req)) for req in required_by.get(field) or ()
                }
                normal_fields.append(
                    (self.aliaser(get_alias(field)), deserializer, required)
                )

        @DeserializationMethodFactory.from_type(cls)
        def factory(
            constraints: Optional[Constraints], validators: Sequence[Validator]
        ) -> DeserializationMethod:
            def method(ctx: DeserializationContext, data: Any) -> Any:
                data = ctx.coercer(dict, data)
                values: Dict[str, Any] = {}
                aliases: List[str] = []
                errors = [] if constraints is None else constraints.errors(data)
                field_errors: Dict[ErrorKey, ValidationError] = OrderedDict()

                for alias, deserialize_field, required in normal_fields:
                    if alias in data:
                        aliases.append(alias)
                        deserialize_field(
                            ctx, data[alias], values, errors, field_errors
                        )
                    elif not required:
                        pass
                    elif required is True:
                        field_errors[alias] = MISSING_PROPERTY
                    else:
                        assert isinstance(required, AbstractSet)
                        requiring = required & data.keys()
                        if requiring:
                            msg = f"missing property (required by {sorted(requiring)})"
                            field_errors[alias] = ValidationError([msg])

                for merged_alias, deserialize_field in merged_fields:
                    merged = {
                        alias: data[alias] for alias in merged_alias if alias in data
                    }
                    aliases.extend(merged)
                    deserialize_field(ctx, merged, values, errors, field_errors)
                if len(data) != len(aliases):
                    remain = data.keys() - set(aliases)
                    for pattern, deserialize_field in pattern_fields:
                        matched = {
                            key: data[key] for key in remain if pattern.match(key)
                        }
                        remain -= matched.keys()
                        deserialize_field(ctx, matched, values, errors, field_errors)
                    if additional_field is not None:
                        additional = {key: data[key] for key in remain}
                        additional_field(ctx, additional, values, errors, field_errors)
                    elif remain and not ctx.additional_properties:
                        for key in remain:
                            field_errors[key] = UNEXPECTED_PROPERTY
                else:
                    for _, deserialize_field in pattern_fields:
                        deserialize_field(ctx, {}, values, errors, field_errors)
                    if additional_field is not None:
                        additional_field(ctx, {}, values, errors, field_errors)
                validators2: Sequence[Validator]
                if validators:
                    init: Dict[str, Any] = {}
                    for init_field in init_vars:
                        if init_field.name in values:
                            init[init_field.name] = values[init_field.name]
                        if (
                            init_field.name not in field_errors
                            and init_field.name in defaults
                        ):
                            init[init_field.name] = defaults[init_field.name]()
                    # Don't keep validators when all dependencies are default
                    validators2 = [
                        v for v in validators if v.dependencies & values.keys()
                    ]
                    if field_errors or errors:
                        error = ValidationError(errors, field_errors)
                        invalid_fields = field_errors.keys() | post_init_modified
                        validators2 = [
                            v
                            for v in validators2
                            if not v.dependencies & invalid_fields
                        ]
                        try:
                            validate(ValidatorMock(cls, values), validators2, **init)
                        except ValidationError as err:
                            error = merge_errors(
                                error, apply_aliaser(err, self.aliaser)
                            )
                        raise error
                elif field_errors or errors:
                    raise ValidationError(errors, field_errors)
                else:
                    validators2, init = ..., ...  # type: ignore # only for linter
                try:
                    res = cls(**values)
                except AssertionError:
                    raise
                except ValidationError as err:
                    raise apply_aliaser(err, self.aliaser)
                except TypeError as err:
                    if str(err).startswith("__init__() got"):
                        raise Unsupported(cls)
                    else:
                        raise ValidationError([str(err)])
                except Exception as err:
                    raise ValidationError([str(err)])
                try:
                    return validate(res, validators2, **init) if validators else res
                except ValidationError as err:
                    raise apply_aliaser(err, self.aliaser)

            return method

        return factory

    def enum(self, cls: Type[Enum]) -> DeserializationMethodFactory:
        deserialize_literal = self.literal([elt.value for elt in cls]).method

        @DeserializationMethodFactory.from_type(cls)
        def factory(
            constraints: Optional[Constraints], validators: Sequence[Validator]
        ) -> DeserializationMethod:
            @with_validators(validators)
            def method(ctx: DeserializationContext, data: Any) -> Any:
                try:
                    result = cls(data)
                except ValueError:
                    result = cls(deserialize_literal(ctx, data))
                if constraints is not None:
                    constraints.validate(data)
                return result

            return method

        return factory

    def literal(self, values: Sequence[Any]) -> DeserializationMethodFactory:
        values_deserializers = [(value, self.method(type(value))) for value in values]

        @DeserializationMethodFactory
        def factory(
            constraints: Optional[Constraints], validators: Sequence[Validator]
        ) -> DeserializationMethod:
            @with_validators(validators)
            def method(ctx: DeserializationContext, data: Any) -> Any:
                if data in values:
                    result = data
                else:
                    for value, deserialize_value in values_deserializers:
                        try:
                            # Literal can contain Enum values which has to be visited
                            if (
                                deserialize_value(ctx, ctx.coercer(type(value), data))
                                == value
                            ):
                                result = value
                                break
                        except Exception:
                            continue
                    else:
                        allowed_values = [
                            value if not isinstance(value, Enum) else value.value
                            for value in values
                        ]
                        raise ValidationError([f"not one of {allowed_values}"])
                if constraints is not None:
                    constraints.validate(result)
                return result

            return method

        return factory

    def mapping(
        self, cls: Type[Mapping], key_type: AnyType, value_type: AnyType
    ) -> DeserializationMethodFactory:
        deserialize_key = self.method(key_type)
        deserialize_value = self.method(value_type)

        @DeserializationMethodFactory
        def factory(
            constraints: Optional[Constraints], validators: Sequence[Validator]
        ) -> DeserializationMethod:
            @with_validators(validators)
            def method(ctx: DeserializationContext, data: Any) -> Any:
                data = ctx.coercer(dict, data)
                items = {}
                item_errors: Dict[ErrorKey, ValidationError] = {}
                for key, value in data.items():
                    assert isinstance(key, str)
                    try:
                        items[deserialize_key(ctx, key)] = deserialize_value(ctx, value)
                    except ValidationError as err:
                        item_errors[key] = err
                errors = () if constraints is None else constraints.errors(data)
                if item_errors or errors:
                    raise ValidationError(errors, item_errors)
                return items if cls is DICT_TYPE else MAPPING_TYPES[cls](items)

            return method

        return factory

    def named_tuple(
        self,
        cls: Type[Tuple],
        types: Mapping[str, AnyType],
        defaults: Mapping[str, Any],
    ) -> DeserializationMethodFactory:
        items_deserializers = [
            (key, self.aliaser(key), self.method(tp)) for key, tp in types.items()
        ]

        @DeserializationMethodFactory.from_type(cls)
        def factory(
            constraints: Optional[Constraints], validators: Sequence[Validator]
        ) -> DeserializationMethod:
            @with_validators(validators)
            def method(ctx: DeserializationContext, data: Any) -> Any:
                data = ctx.coercer(dict, data)
                items: Dict[str, Any] = {}
                item_errors: Dict[ErrorKey, ValidationError] = {}
                for key, alias, deserialize_item in items_deserializers:
                    if key in data:
                        try:
                            items[key] = deserialize_item(ctx, data[alias])
                        except ValidationError as err:
                            if key not in defaults or not ctx.default_fallback:
                                item_errors[alias] = err
                    elif key not in defaults:
                        item_errors[alias] = MISSING_PROPERTY

                if not ctx.additional_properties:
                    for key in sorted(data.keys() - defaults.keys() - items.keys()):
                        item_errors[key] = UNEXPECTED_PROPERTY
                errors = () if constraints is None else constraints.errors(data)
                if item_errors or errors:
                    raise ValidationError(errors, item_errors)
                return cls(**items)

            return method

        return factory

    def new_type(
        self, tp: AnyType, super_type: AnyType
    ) -> DeserializationMethodFactory:
        return self.visit(super_type).merge(get_constraints(tp), get_validators(tp))

    def primitive(self, cls: Type) -> DeserializationMethodFactory:
        @DeserializationMethodFactory
        def factory(
            constraints: Optional[Constraints], validators: Sequence[Validator]
        ) -> DeserializationMethod:
            if constraints is None:
                method = lambda ctx, data: ctx.coercer(cls, data)  # noqa: E731
            else:

                def method(ctx: DeserializationContext, data: Any) -> Any:
                    data = ctx.coercer(cls, data)
                    if constraints is not None and data is not None:
                        constraints.validate(data)
                    return data

            return with_validators(validators)(method)

        return factory

    def subprimitive(self, cls: Type, superclass: Type) -> DeserializationMethodFactory:
        super_factory = self.primitive(superclass)

        @DeserializationMethodFactory.from_type(cls)
        def factory(
            constraints: Optional[Constraints], validators: Sequence[Validator]
        ) -> DeserializationMethod:
            method = super_factory.merge(constraints, validators).method
            return lambda ctx, data: cls(method(ctx, data))

        return factory

    def tuple(self, types: Sequence[AnyType]) -> DeserializationMethodFactory:
        tuple_constraints = ArrayConstraints(min_items=len(types), max_items=len(types))
        elts_deserializers = [self.method(cls) for cls in types]

        @DeserializationMethodFactory
        def factory(
            constraints: Optional[Constraints], validators: Sequence[Validator]
        ) -> DeserializationMethod:
            @with_validators(validators)
            def method(ctx: DeserializationContext, data: Any) -> Any:
                data = ctx.coercer(list, data)
                if len(data) != len(types):
                    tuple_constraints.validate(data)
                elts: List[Any] = []
                elt_errors: Dict[ErrorKey, ValidationError] = {}
                for i, (deserialize_elt, elt) in enumerate(
                    zip(elts_deserializers, data)
                ):
                    try:
                        elts.append(deserialize_elt(ctx, elt))
                    except ValidationError as err:
                        elt_errors[i] = err
                errors = () if constraints is None else constraints.errors(data)
                if elt_errors or errors:
                    raise ValidationError(errors, elt_errors)
                return tuple(elts)

            return method

        return factory

    def typed_dict(
        self, cls: Type, keys: Mapping[str, AnyType], total: bool
    ) -> DeserializationMethodFactory:
        items_deserializers = {key: self.method(tp) for key, tp in keys.items()}

        @DeserializationMethodFactory.from_type(cls)
        def factory(
            constraints: Optional[Constraints], validators: Sequence[Validator]
        ) -> DeserializationMethod:
            @with_validators(validators)
            def method(ctx: DeserializationContext, data: Any) -> Any:
                data = ctx.coercer(dict, data)
                items: Dict[str, Any] = {}
                key_count = 0
                item_errors: Dict[ErrorKey, ValidationError] = {}
                for key, value in data.items():
                    if key in items_deserializers:
                        key_count += 1
                        try:
                            items[key] = items_deserializers[key](ctx, value)
                        except ValidationError as err:
                            item_errors[key] = err
                    else:
                        items[key] = value
                if total and key_count != len(keys):
                    for key in keys:
                        if key not in data:
                            item_errors[key] = MISSING_PROPERTY
                errors = () if constraints is None else constraints.errors(data)
                if item_errors or errors:
                    raise ValidationError(errors, item_errors)
                return items

            return method

        return factory

    def union(self, alternatives: Sequence[AnyType]) -> DeserializationMethodFactory:
        factories = [self.visit(cls) for cls in filter_skipped(alternatives)]
        optional = NoneType in alternatives

        @DeserializationMethodFactory
        def factory(
            constraints: Optional[Constraints], validators: Sequence[Validator]
        ) -> DeserializationMethod:
            alt_deserializers = [
                fact.merge(constraints, validators).method for fact in factories
            ]

            def method(ctx: DeserializationContext, data: Any) -> Any:
                # Optional optimization
                if data is None and optional:
                    return None
                error: Optional[ValidationError] = None
                for deserialize_alt in alt_deserializers:
                    try:
                        return deserialize_alt(ctx, data)
                    except ValidationError as err:
                        error = merge_errors(error, err)
                else:
                    if error is None:  # empty union
                        return data
                    else:
                        raise error

            return method

        return factory

    def visit_conversion(
        self, tp: AnyType, conversion: Deserialization, dynamic: bool
    ) -> DeserializationMethodFactory:
        assert conversion
        cls = get_origin_or_type(tp)
        factories = [
            (
                self.visit_with_conversions(
                    self._update_generic_args(tp, conv),
                    handle_container_conversions(
                        conv.source, conv.sub_conversions, self._conversions, dynamic
                    ),
                ),
                cast(Converter, conv.converter),
                (conv.additional_properties, conv.coercion, conv.default_fallback),
            )
            for conv in conversion
        ]

        @DeserializationMethodFactory.from_type(cls)
        def factory(
            constraints: Optional[Constraints], validators: Sequence[Validator]
        ) -> DeserializationMethod:
            alt_deserializers = [
                (fact.merge(constraints, validators).method, converter, conv_ctx)
                for fact, converter, conv_ctx in factories
            ]
            if len(alt_deserializers) == 1:
                deserialize_alt, converter, conv_ctx = alt_deserializers[0]
                if converter is identity:
                    method = deserialize_alt
                else:

                    def method(ctx: DeserializationContext, data: Any) -> Any:
                        try:
                            return converter(
                                deserialize_alt(ctx.merge(*conv_ctx), data)
                            )
                        except (ValidationError, AssertionError):
                            raise
                        except Exception as err:
                            raise ValidationError([str(err)])

            else:

                def method(ctx: DeserializationContext, data: Any) -> Any:
                    error: Optional[ValidationError] = None
                    for deserialize_alt, converter, conv_ctx in alt_deserializers:
                        try:
                            value = deserialize_alt(ctx.merge(*conv_ctx), data)
                            break
                        except ValidationError as err:
                            error = merge_errors(error, err)
                    else:
                        assert error is not None
                        raise error
                    try:
                        return converter(value)
                    except (ValidationError, AssertionError):
                        raise
                    except Exception as err:
                        raise ValidationError([str(err)])

            return method

        return factory


@cache
def get_method(
    tp: AnyType, conversions: Optional[Conversions], aliaser: Aliaser
) -> DeserializationMethod:
    factory = DeserializationMethodVisitor(aliaser).visit_with_conversions(
        tp, conversions
    )
    return factory.method


@overload
def deserialize(
    tp: Type[T],
    data: Any,
    *,
    conversions: Conversions = None,
    aliaser: Aliaser = None,
    additional_properties: bool = None,
    coercion: Coercion = None,
    default_fallback: bool = None,
) -> T:
    ...


@overload
def deserialize(
    tp: AnyType,
    data: Any,
    *,
    conversions: Conversions = None,
    aliaser: Aliaser = None,
    additional_properties: bool = None,
    coercion: Coercion = None,
    default_fallback: bool = None,
) -> Any:
    ...


def deserialize(
    tp: AnyType,
    data: Any,
    *,
    conversions: Conversions = None,
    aliaser: Aliaser = None,
    additional_properties: bool = None,
    coercion: Coercion = None,
    default_fallback: bool = None,
):
    if additional_properties is None:
        additional_properties = settings.additional_properties
    if coercion is None:
        coercion = settings.coercion
    if default_fallback is None:
        default_fallback = settings.default_fallback
    if aliaser is None:
        aliaser = settings.aliaser()
    ctx = DeserializationContext(additional_properties, coercion, default_fallback)
    if isinstance(conversions, Collection):
        conversions = tuple(conversions)  # Make it hashable
    return get_method(tp, conversions, aliaser)(ctx, data)



from collections import defaultdict
from dataclasses import Field
from functools import wraps
from inspect import Parameter, isgeneratorfunction, signature
from types import MethodType
from typing import (
    AbstractSet,
    Any,
    Callable,
    Collection,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    overload,
)

from apischema.conversions.dataclass_models import get_model_origin, has_model_origin
from apischema.types import AnyType
from apischema.typing import get_type_hints
from apischema.utils import get_origin_or_type, is_method, method_class
from apischema.validation.dependencies import find_all_dependencies
from apischema.validation.errors import (
    Error,
    FieldPath,
    ValidationError,
    merge_errors,
    yield_to_raise,
)
from apischema.validation.mock import NonTrivialDependency

_validators: Dict[Type, List["Validator"]] = defaultdict(list)


def get_validators(tp: AnyType) -> Sequence["Validator"]:
    validators = []
    if hasattr(tp, "__mro__"):
        for sub_cls in tp.__mro__:
            validators.extend(_validators[sub_cls])
    else:
        validators.extend(_validators[tp])
    if has_model_origin(tp):
        validators.extend(get_validators(get_model_origin(tp)))
    return validators


class Discard(Exception):
    def __init__(self, fields: AbstractSet[Field], error: ValidationError):
        self.fields = fields
        self.error = error


class Validator:
    def __init__(
        self,
        func: Callable,
        field: Field = None,
        discard: Union[Field, Collection[Field]] = None,
    ):
        wraps(func)(self)
        self.func = func
        self.field = field
        # Cannot use field.name because fields are not yet initialized with __set_name__
        if field is not None and discard is None:
            self.discard: AbstractSet[Field] = {field}
        elif isinstance(discard, Field):
            self.discard = {discard}
        else:
            self.discard = set(discard or ())
        self.dependencies: AbstractSet[str] = set()
        validate = func
        try:
            parameters = signature(func).parameters
        except ValueError:
            self.params: AbstractSet[str] = set()
        else:
            if not parameters:
                raise TypeError("Validator must have at least one parameter")
            if any(p.kind == Parameter.VAR_KEYWORD for p in parameters.values()):
                raise TypeError("Validator cannot have variadic keyword parameter")
            if any(p.kind == Parameter.VAR_POSITIONAL for p in parameters.values()):
                raise TypeError("Validator cannot have variadic positional parameter")
            self.params = set(list(parameters)[1:])
        if isgeneratorfunction(func):
            validate = yield_to_raise(validate)
        if self.field is not None:
            wrapped_field = validate

            def validate(__obj, **kwargs):
                try:
                    wrapped_field(__obj, **kwargs)
                except ValidationError as err:
                    raise ValidationError(children={FieldPath(field): err})

        if self.discard:
            wrapped_discard = validate

            def validate(__obj, **kwargs):
                try:
                    wrapped_discard(__obj, **kwargs)
                except ValidationError as err:
                    raise Discard(self.discard, err)

        self.validate: Callable[..., Iterator[Error]] = validate
        self._registered = False

    def __get__(self, instance, owner):
        return self if instance is None else MethodType(self.func, instance)

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def _register(self, owner: Type):
        if self._registered:
            raise RuntimeError("Validator already registered")
        self.dependencies = find_all_dependencies(owner, self.func) | self.params
        _validators[owner].append(self)
        self._registered = True

    def __set_name__(self, owner, name):
        self._register(owner)


T = TypeVar("T")


def validate(__obj: T, __validators: Iterable[Validator] = None, **kwargs) -> T:
    if __validators is None:
        __validators = get_validators(type(__obj))
    error: Optional[ValidationError] = None
    __validators = iter(__validators)
    while True:
        for validator in __validators:
            try:
                if kwargs and validator.params != kwargs.keys():
                    assert all(k in kwargs for k in validator.params)
                    validator.validate(
                        __obj, **{k: kwargs[k] for k in validator.params}
                    )
                else:
                    validator.validate(__obj, **kwargs)
            except ValidationError as err:
                error = merge_errors(error, err)
            except Discard as err:
                error = merge_errors(error, err.error)
                discarded = {f.name for f in err.fields}
                __validators = iter(
                    v for v in __validators if not discarded & v.dependencies
                )
                break
            except NonTrivialDependency as exc:
                exc.validator = validator
                raise
            except AssertionError:
                raise
            except Exception as err:
                error = merge_errors(error, ValidationError([str(err)]))
        else:
            break
    if error is not None:
        raise error
    return __obj


V = TypeVar("V", bound=Callable)


@overload
def validator(func: V) -> V:
    ...


@overload
def validator(
    field: Any = None, *, discard: Any = None, owner: Type = None
) -> Callable[[V], V]:
    ...


def validator(arg=None, *, discard=None, owner=None):
    if arg is None:
        return lambda func: Validator(func, None, discard)
    if callable(arg):
        validator_ = Validator(arg, None, None)
        if is_method(arg) and method_class(arg) is None:
            return validator_
        if is_method(arg):
            if owner is None:
                owner = method_class(arg)
        if owner is None:
            try:
                first_param = next(iter(signature(arg).parameters))
                owner = get_origin_or_type(get_type_hints(arg)[first_param])
            except Exception:
                raise ValueError("Validator first parameter must be typed")
        validator_._register(owner)
        return validator_
    field = arg
    if not isinstance(field, Field):
        raise TypeError(
            "Validator argument must be a field declared with `... = field(...)`"
        )
    if discard is not None:
        if not isinstance(discard, Field) or not (
            isinstance(discard, Collection)
            and all(isinstance(f, Field) for f in discard)
        ):
            raise TypeError("discard must be a field or a collection of fields")
    return lambda func: Validator(func, field, discard)
